/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package actor.components;

import gameapp.GameApp;

import java.util.LinkedList;
import java.util.List;

import javax.vecmath.Quat4f;
import javax.vecmath.Vector3f;

import logging.Logger;

import org.lwjgl.util.vector.Matrix4f;
import org.w3c.dom.Node;

import util.MathUtil;
import util.Pair;
import util.Pose;
import util.XMLUtil;
import actor.ActorComponent;

import com.bulletphysics.collision.dispatch.CollisionFlags;
import com.bulletphysics.collision.dispatch.CollisionObject;
import com.bulletphysics.collision.shapes.BoxShape;
import com.bulletphysics.collision.shapes.CollisionShape;
import com.bulletphysics.collision.shapes.CompoundShape;
import com.bulletphysics.collision.shapes.SphereShape;
import com.bulletphysics.dynamics.RigidBody;
import com.bulletphysics.dynamics.RigidBodyConstructionInfo;
import com.bulletphysics.linearmath.MotionState;
import com.bulletphysics.linearmath.Transform;
/**
 * Ein Actor mit einer PhysicsComponent wird im PhysicsLayer als dynamisches
 * Objekt behandelt.
 * @author nico3000
 */
public class PhysicsComponent extends ActorComponent {
    /**
     * Die ID der Komponente.
     */
    public static final int ID = 0xa8b369f5;
    
    private float mass = 0.0f;
    
    private RigidBody body = null;
    
    private final List<Pair<CollisionShape, Transform>> shapes = new LinkedList<>();
    
    private boolean kinematic = false;

    @Override
    protected int getComponentId() {
        return ID;
    }

    @Override
    protected boolean init(Node data) {
        Float customMass = null;
        while(data != null) {
            switch(data.getNodeName()) {
                case "Box":
                    this.createBox(data);
                    break;
                case "Sphere":
                    this.createSphere(data);
                    break;
                case "Mass":
                    customMass = XMLUtil.getAttributeAsFloat(data, "value", 0);
                    break;
                case "Property":
                	switch (XMLUtil.getAttributeAsString(data, "name")) {
                		case "kinematic":
                			this.kinematic = XMLUtil.getAttributeAsBoolean(data, "value", false);
                			Logger.INSTANCE.info(this.getOwner().getName() + " is kinematic: " + this.kinematic);
                			break;
                	}
                case "#comment":
                case "#text":
                    // nothing to do
                    break;
                default:
                    Logger.INSTANCE.error("Node " + data.getNodeName() + " was ignored");
                    break;
            }
            data = data.getNextSibling();
        }
        if(this.shapes.isEmpty()) {
            Logger.INSTANCE.error("empty PhysicsComponent");
            return false;
        } else {
            if(customMass != null) {
                this.mass = customMass;
            }
            return true;
        }
    }

    @Override
    protected void postInit() {
        TransformComponent comp = (TransformComponent)this.getOwner().getComponent(TransformComponent.ID);
        if(comp != null) {
            CompoundShape compound = new CompoundShape();
            for(Pair<CollisionShape, Transform> pair : this.shapes) {
                compound.addChildShape(pair.right, pair.left);
            }
            javax.vecmath.Vector3f localInertia = new javax.vecmath.Vector3f();
            compound.calculateLocalInertia(this.mass, localInertia);            
            RigidBodyConstructionInfo info = new RigidBodyConstructionInfo(this.mass, new MotionStateCallback(comp.getPose()), compound, localInertia);
            
            info.angularSleepingThreshold = 1e-1f;
            info.linearSleepingThreshold = 1e-1f;
            if(this.shapes.size() == 1 && this.shapes.get(0).left instanceof SphereShape) {
                info.restitution = 0.8f;
            } else {
                info.restitution = 0.3f;
            }
            info.angularDamping = 0.1f;
            this.body = new RigidBody(info);
            if (this.kinematic) {
            	this.body.setCollisionFlags(this.body.getCollisionFlags() | CollisionFlags.KINEMATIC_OBJECT);
            	this.body.setActivationState(CollisionObject.DISABLE_DEACTIVATION);
            }
            this.body.setUserPointer(this.getOwner());
//			body.setLinearVelocity(new Vector3f(2.0f, 0.0f,0.0f));
//			body.setAngularVelocity(new Vector3f(0f, 1f, 0f));
//			
//			body.setCcdMotionThreshold(1f);
//			body.setCcdSweptSphereRadius(0.2f);
            GameApp.getPhysicsLayer().addRigidBodyToWorld(this.body);
            
        } else {
            Logger.INSTANCE.warning("actor with PhysicsComponent but no TransformComponent");
        }
    }
    
    /**
     * Erzeugt eine simple Box als Geometrie.
     * @param shapeData Der <tt>Box</tt> Node
     */
    private void createBox(Node shapeData) {
    	org.lwjgl.util.vector.Vector3f lwjglVec = XMLUtil.getAttributesAsVector(XMLUtil.getChildByType(shapeData, "Size"), "x", "y", "z", null);
        Vector3f size = new Vector3f(lwjglVec.x, lwjglVec.y, lwjglVec.z);
        float density = XMLUtil.getAttributeAsFloat(XMLUtil.getChildByType(shapeData, "Density"), "value", 0.0f);
        
        float volume = size.x * size.y * size.z;
        this.mass += volume * density;
        CollisionShape boxShape = new BoxShape(new javax.vecmath.Vector3f(0.5f * size.x, 0.5f * size.y, 0.5f * size.z));
        this.shapes.add(new Pair<>(boxShape, this.getTransform(null, shapeData)));
    }
    
    /**
     * Erzeugt eine simple Kugel als Geometrie.
     * @param shapeData Der <tt>Sphere</tt> Node
     */
    private void createSphere(Node shapeData) {
        float radius = XMLUtil.getAttributeAsFloat(XMLUtil.getChildByType(shapeData, "Size"), "radius", 0.0f);
        float density = XMLUtil.getAttributeAsFloat(XMLUtil.getChildByType(shapeData, "Density"), "value", 0.0f);
        
        float volume = 4.0f / 3.0f * MathUtil.PI * radius * radius * radius;
        this.mass += volume * density;
        CollisionShape sphereShape = new SphereShape(radius);
        this.shapes.add(new Pair<>(sphereShape, this.getTransform(null, shapeData)));
    }
    
    private Transform getTransform(Transform dest, Node shapeData) {
        if(dest == null) {
            dest = new Transform();
        }
        org.lwjgl.util.vector.Vector3f lwjglPositionVec = XMLUtil.getAttributesAsVector(XMLUtil.getChildByType(shapeData, "Position"), "x", "y", "z", null);
        org.lwjgl.util.vector.Vector3f lwjglRotVec = XMLUtil.getAttributesAsVector(XMLUtil.getChildByType(shapeData, "Rotation"), "yaw", "pitch", "roll", null);
        MathUtil.toRadians(lwjglRotVec);
        Matrix4f model = MathUtil.generateModel(null, null, MathUtil.quaternionFromYPR(null, lwjglRotVec.x, lwjglRotVec.y, lwjglRotVec.z), lwjglPositionVec, 1.0f);
        dest.set(MathUtil.lwjgl2vecmath(null, model));
        return dest;
    }

    @Override
    protected void destroy() {
        GameApp.getPhysicsLayer().removeRigidBodyFromWorld(this.body);
        this.shapes.clear();
    }
    
    /**
     * 
     */
    private class MotionStateCallback extends MotionState {
        /**
         * Die Pose, des Actors aus der TransformComponent.
         */
        private final Pose pose;

        /**
         * 
         * @param pose 
         */
        public MotionStateCallback(Pose pose) {
            this.pose = pose;
        }

        @Override
        public Transform getWorldTransform(Transform out) {
            return MathUtil.lwjgl2jbullet(out, this.pose.getPosition(), this.pose.getRotation());
        }

        @Override
        public void setWorldTransform(Transform worldTrans) {
            this.pose.setPosition(worldTrans.origin.x, worldTrans.origin.y, worldTrans.origin.z);
            this.pose.setRotation(MathUtil.vecmath2lwjgl(null, worldTrans.getRotation(new Quat4f())));
        }
    }
}
