/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package gameapp;

import graphics.geometry.IndexBuffer;
import java.nio.ByteBuffer;
import logging.Logger;
import gameapp.GameApp;
import org.lwjgl.BufferUtils;
import org.lwjgl.LWJGLException;
import org.lwjgl.opengl.Display;
import org.lwjgl.opengl.DisplayMode;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL30;
import org.lwjgl.opengl.GL31;
import resource.ResHandle;
import resource.Resource;

/**
 * Stellt die Schnittstelle zur Grafikprogrammierung bereit.
 * @author nico3000
 */
public class GraphicsLayer {
    /**
     * Colorbufferbit fuer <tt>clear()</tt>.
     */
    public static final int MASK_COLOR = GL11.GL_COLOR_BUFFER_BIT;
    /**
     * Depthbufferbit fuer <tt>clear()</tt>.
     */
    public static final int MASK_DEPTH = GL11.GL_DEPTH_BUFFER_BIT;
    /**
     * Stencilbufferbit fuer <tt>clear()</tt>.
     */
    public static final int MASK_STENCIL = GL11.GL_STENCIL_BUFFER_BIT;
    
    /**
     * Die FPS des letzten Frames.
     */
    private long fps = 0;
    /**
     * Zaehler zur FPS Berechnung.
     */
    private long frameCount = 0;
    /**
     * Akkumulator fuer FPS Berechnung.
     */
    private long fpsMillisAccumulator = 0;
    
    /**
     * Initialisiert das Graphicsdevice.
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean init() {
        try {
            this.setPreferredDisplayMode();
            Display.create();
        } catch (LWJGLException ex) {
            Logger.INSTANCE.error(ex.getMessage());
            return false;
        }
        GL31.glPrimitiveRestartIndex(IndexBuffer.RESTART);
        GL11.glEnable(GL31.GL_PRIMITIVE_RESTART);
        GL11.glEnable(GL11.GL_CULL_FACE);
        GL11.glEnable(GL11.GL_DEPTH_TEST);
        GL11.glEnable(GL11.GL_BLEND);
        GL11.glFrontFace(GL11.GL_CCW);
        GL11.glBlendFunc(GL11.GL_SRC_ALPHA, GL11.GL_ONE_MINUS_SRC_ALPHA);
        return true;
    }
    
    /**
     * Gibt das gesamte Device wieder frei.
     */
    public void destroy() {
        Display.destroy();
    }
    
    /**
     * Setzt die bevorzugten Anzeigeeinstellungen (Aufloesung, Bildfrequenz
     * Fenster / Vollbild)
     * @throws LWJGLException Bei OpenGL Fehlern
     */
    private void setPreferredDisplayMode() throws LWJGLException {
        boolean windowed = GameApp.getConfig().getAsBoolean("bWindowed");
        int resX = GameApp.getConfig().getAsInt("iResolutionX");
        int resY = GameApp.getConfig().getAsInt("iResolutionY");
        int frequency = GameApp.getConfig().getAsInt("iFrequency");
        Display.setVSyncEnabled(GameApp.getConfig().getAsBoolean("bVerticalSync"));
        if(windowed) {
            Display.setFullscreen(false);
            Display.setDisplayMode(new DisplayMode(resX, resY));
        } else {
            DisplayMode target = null;
            for(DisplayMode mode : Display.getAvailableDisplayModes()) {
                if(mode.getWidth() == resX && mode.getHeight() == resY) {
                    Display.setVSyncEnabled(windowed);
                    if(target == null || mode.getFrequency() == frequency) {
                        target = mode;
                    }
                }
            }
            if(target != null) {
                Display.setDisplayModeAndFullscreen(target);
            } else {
                Display.setFullscreen(false);
                Display.setDisplayMode(new DisplayMode(640, 480));
            }
        }
    }
    
    /**
     * Setzt die Farbe, mit der der Colorbuffer gecleart werden soll.
     * @param r Rotwert
     * @param g Gruenwert
     * @param b Blauwert
     * @param a Alphawert
     */
    public void setClearColor(float r, float g, float b, float a) {
        GL11.glClearColor(r, g, b, a);
    }
    
    /**
     * Setzt die Werte mit denen der Depth- und Stencilbuffer gecleart werden
     * soll.
     * @param depth Der Wert des Depthbuffers
     * @param stencil Der Wert des Stencilbuffers
     */
    public void setDepthStencilClearValues(float depth, int stencil) {
        GL11.glClearDepth(depth);
        GL11.glClearStencil(stencil);
    }
    
    /**
     * Setzt den Titel des GL Fensters.
     * @param title Der neue Titel
     */
    public void setWindowTitle(String title) {
        Display.setTitle(title);
    }
    
    /**
     * Liefert true, falls der Benutzer seit dem letzten Aufruf das Fenster
     * schliessen wollte.
     * @return <tt>true</tt>, falls der Benutzer das Fenster schliessen wollte
     */
    public boolean isWindowCloseRequested() {
        return Display.isCloseRequested();
    }
    
    /**
     * Fuellt den Backbuffer mit der Clearcolor.
     * @param mask Maske der zu clearenden Buffer
     */
    public void clear(int mask) {
        GL11.glClear(mask);
    }
    
    /**
     * Tauscht Front- und Backbuffer und aktualisiert die FPS.
     * @param deltaMillis Die Dauer des Frames in Millisekunden (Realtimer)
     */
    public void update(long deltaMillis) {
        ++this.frameCount;
        if((this.fpsMillisAccumulator += deltaMillis) > 200) {
            this.fps = 1000 * this.frameCount / this.fpsMillisAccumulator;
            this.fpsMillisAccumulator = 0;
            this.frameCount = 0;
        }
        Display.update();
    }

    /**
     * Liefert die Frames pro Sekunde.
     * @return FPS
     */
    public long getFps() {
        return this.fps;
    }

    /**
     * Die Breite des Fensters in Pixel.
     * @return Die Breite
     */
    public int getWidth() {
        return Display.getWidth();
    }

    /**
     * Die Hoehe des Fensters in Pixel.
     * @return Die Hoehe
     */
    public int getHeight() {
        return Display.getHeight();
    }
    
    /**
     * Laedt eine Resource direkt in einen ByteBuffer.
     * @param resource Die Resource
     * @return Der ByteBuffer oder <tt>null</tt>
     */
    public ByteBuffer resource2ByteBuffer(Resource resource) {
        ResHandle handle = GameApp.getResCache().getHandle(resource);
        if(handle == null) {
            return null;
        }
        ByteBuffer buffer = BufferUtils.createByteBuffer(handle.getSize());
        buffer.put(handle.getBuffer());
        buffer.position(0);
        handle.unlock();
        return buffer;
    }
    
    /**
     * Prueft auf Fehler in de GL und gibt diese aus.
     * @return <tt>true</tt>, falls keine Fehler
     */
    public boolean isErrorFree() {
        int error = GL11.glGetError();
        switch(error) {
            case GL11.GL_NO_ERROR:
                return true;
            case GL11.GL_INVALID_ENUM:
                Logger.INSTANCE.error("GL_INVALID_ENUM");
                break;
            case GL11.GL_INVALID_OPERATION:
                Logger.INSTANCE.error("GL_INVALID_OPERATION");
                break;
            case GL11.GL_INVALID_VALUE:
                Logger.INSTANCE.error("GL_INVALID_VALUE");
                break;
            case GL30.GL_INVALID_FRAMEBUFFER_OPERATION:
                Logger.INSTANCE.error("GL_INVALID_FRAMEBUFFER_OPERATION");
                break;
            case GL11.GL_OUT_OF_MEMORY:
                Logger.INSTANCE.error("GL_OUT_OF_MEMORY");
                break;
            case GL11.GL_STACK_OVERFLOW:
                Logger.INSTANCE.error("GL_STACK_OVERFLOW");
                break;
            case GL11.GL_STACK_UNDERFLOW:
                Logger.INSTANCE.error("GL_STACK_UNDERFLOW");
                break;
        }
        return false;
    }
}
