/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package graphics.geometry;

import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import logging.Logger;
import mesh.IndexBufferInterval;
import mesh.material.Material;
import mesh.material.MaterialLibrary;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL30;
import org.lwjgl.util.vector.Vector3f;

/**
 * Eine Geometrie ist eine Veknuepfung von einem Index- und einem Vertexbuffer.
 * @author nico3000
 */
public class Geometry {   
    /**
     * Alle erzeugten Instanzen der Klasse Geometry indiziert nach ihrer
     * Resource.
     */
    private static final Map<String, Geometry> INSTANCES = new HashMap<>();
    
    /**
     * Liefert eine Geometry und erzeugt sie, falls sie nicht existiert. Diese
     * Methode initialisiert die Geometry <i>nicht</i>!
     * @param resource Die Resource
     * @return Die Geometry
     */
    public static Geometry createGeometry(String resource) {
        Geometry geometry = INSTANCES.get(resource);
        if(geometry == null) {
            INSTANCES.put(resource, geometry = new Geometry(resource));
        }
        ++geometry.ref;
        return geometry;
    }
    
    /**
     * Liefert eine Geometry und erzeugt sie, falls sie nicht existiert. Diese
     * Methode initialisiert die Geometry mit den uebergebenen Buffern.
     * @param resource Die Resource
     * @return Die Geometry
     */
    public static Geometry initGeometry(VertexBuffer vertexBuffer, IndexBuffer indexBuffer) {
        Geometry geometry = new Geometry("generic_geometry_" + vertexBuffer.hashCode() + "_" + indexBuffer.hashCode());
        if(!geometry.init(indexBuffer, vertexBuffer)) {
            Logger.INSTANCE.error("Failed to initialize geometry!");
            return null;
        } else {
            ++geometry.ref;
            INSTANCES.put(geometry.getResource(), geometry);
            return geometry;
        }
    }
    
    /**
     * Gibt den Status der bekannten Geometrien zurueck.
     * @return Der Status
     */
    public static String getDebugStatus() {
        String geometries = "";
        for(Geometry geometry : INSTANCES.values()) {
            if(geometry.isInitialized()) {
                geometries += "\n- " + geometry.resource;
            }
        }
        return String.format("Active Geometry instances: %s", geometries.isEmpty() ? "none" : geometries);
    }
    
    /**
     * Die ID der Vertex array obejct.
     */
    private int id = 0;
    /**
     * Reference Counter.
     */
    private int ref = 0;
    /**
     * Die Reosurce der Geometrie.
     */
    private final String resource;
    /**
     * Der Indexbuffer.
     */
    private IndexBuffer indexBuffer = null;
    /**
     * Der Vertexbuffer.
     */
    private VertexBuffer vertexBuffer = null;
    /**
     * Die Indexbuffer Intervalle dieser Geometrie mit zugehoeriger
     * MaterilLibrary und Material.
     */
    private final Map<Material, IndexBufferInterval> intervals = new HashMap<>();
    /**
     * Die Standard MaterialLibrary auf die sich die Materialien der Indexbuffer
     * Intervalle beziehen.
     */
    private Set<MaterialLibrary> referencedLibraries = new HashSet<>();
    
    
    /**
     * Privater Konstruktor, um Erzeugung zu verhindern.
     */
    private Geometry(String resource) {
        this.resource = resource;
    }

    /**
     * Liefert die Resource dieser Geometry.
     * @return Die Resource
     */
    public String getResource() {
        return this.resource;
    }
    
    /**
     * Ordnet einem Material ein Indexbuffer Intervall zu.
     * @param interval Das Intervall
     * @param libraryName Der Name der MaterialLibrary
     * @param materialName Der Name des Materials
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean assignIntervalToMaterial(IndexBufferInterval interval, String libraryName, String materialName) {
        MaterialLibrary library = MaterialLibrary.getLibrary(libraryName);
        if(library == null) {
            Logger.INSTANCE.warning("Unknown MaterialLibrary %s for geometry %s.", libraryName, this.resource);
            return false;
        }
        Material material = library.getMaterialByName(materialName);
        if(material == null) {
            Logger.INSTANCE.warning("Unknwon material %s in MaterialLibrary %s for geometry %s.", materialName, libraryName, this.resource);
            return false;
        }
        if(this.intervals.containsKey(material)) {
            Logger.INSTANCE.warning("Overwriting IndexBufferInterval for Material %s in geometry %s.", materialName, this.resource);
            material.release();
        }
        this.intervals.put(material, interval);
        if(!this.referencedLibraries.add(library)) {
            library.release();
        }
        return true;
    }
    
    /**
     * Liefert die Standard Materialien und die zugehörigen Indexbuffer
     * Intervalle dieser Geometry.
     * @return Die Materialien und Intervalle
     */
    public Iterable<Map.Entry<Material, IndexBufferInterval>> getIntervals() {
        return this.intervals.entrySet();
    }
    
    /**
     * Initialisiert eine Geometry mit einem Index- und Vertexbuffer.
     * @param indexBuffer Der Indexbuffer
     * @param vertexBuffer Der Vertexnuffer
     * @return <tt>true</tt>, falls eroglreich
     */
    public boolean init(IndexBuffer indexBuffer, VertexBuffer vertexBuffer) {
        if(this.isInitialized()) {
            Logger.INSTANCE.warning("Reinitialiazing geometry %s.", this.resource);
            this.destroy();
        }
        this.indexBuffer = indexBuffer;
        this.vertexBuffer = vertexBuffer;
        this.id = GL30.glGenVertexArrays();
        if(this.id < 1) {
            Logger.INSTANCE.error("Generation of VertexArray failed!");
            return false;
        } else {
            this.bind();
            this.indexBuffer.bind();
            this.vertexBuffer.bind();
            if(!this.vertexBuffer.bindAttributes()) {
                Logger.INSTANCE.warning("Geometry %s has no vertex attributes!", this.resource);
            }
            GL30.glBindVertexArray(0);
            return true;
        }
    }
    
    /**
     * Bindet die Geometrie und kann anschliessend mit <tt>draw()</tt> 
     * gezeichnet werden.
     */
    public void bind() {
        GL30.glBindVertexArray(this.id);
    }
    
    /**
     * Gibt diese Instanz wieder frei. Wurde sie ueberall wieder freigegeben,
     * wird sie anschliessend zerstoert.
     */
    public void release() {
        if(this.ref == 0) {
            Logger.INSTANCE.warning("Releasing of geometry %s not possible because reference counter is already 0.");
        } else if(--this.ref == 0) {
            this.destroy();
        }
    }
    
    /**
     * Zerstoert die mit dieser Geometry verbundenen Daten.
     */
    private void destroy() {
        //Logger.INSTANCE.info("Now destroying geometry %s.", this.resource);
        if(this.vertexBuffer != null) {
            this.vertexBuffer.destroy();
            this.vertexBuffer = null;
        }
        if(this.indexBuffer != null) {
            this.indexBuffer.destroy();
            this.indexBuffer = null;
        }
        for(Material material : this.intervals.keySet()) {
            material.release();
        }
        this.intervals.clear();
        for(MaterialLibrary library : this.referencedLibraries) {
            library.release();
        }
        if(this.id > 0) {
            GL30.glDeleteVertexArrays(this.id);
            this.id = 0;
        }
    }
    
    /**
     * Zeichnet die Geometrie. Dazu muss sie vorher mit <tt>bind()</tt> gebunden
     * werden.
     */
    public void render() {
        GL11.glDrawElements(this.indexBuffer.getTopology(), this.indexBuffer.getCount(), GL11.GL_UNSIGNED_INT, 0);       
    }
    
    /**
     * Zeichnet einen Teil der Geometrie. Dazu muss sie vorher mit
     * <tt>bind()</tt> gebunden werden.
     */
    public void drawInterval(int startIndex, int indexCount) {
        GL11.glDrawElements(this.indexBuffer.getTopology(), indexCount, GL11.GL_UNSIGNED_INT, 4 * startIndex);
    }
    
    /**
     * Prueft, ob die Geometry bereits initialisiert wurde.
     * @return <tt>true</tt>, falls sie bereits initialisiert wurde
     */
    public boolean isInitialized() {
        return this.id != 0;
    }
    
    public static boolean calculateNormals(FloatBuffer vertices, int posOffset, int normalOffset, int stride, IntBuffer indices, int topology) {
        if(indices.remaining() < 3) {
            Logger.INSTANCE.warning("Cannot compute normals.");
            return false;
        }
        final Vector3f p0 = new Vector3f();
        final Vector3f p1 = new Vector3f();
        final Vector3f p2 = new Vector3f();
        final Vector3f n = new Vector3f();
        final Vector3f p01 = new Vector3f();
        final Vector3f p02 = new Vector3f();
        Set<Integer> touched = new TreeSet<>();
        boolean done = false;
        int i[] = { indices.get(), indices.get(), indices.get(), };
        while(!done) {
            boolean valid = true;
            switch(topology) {
                case GL11.GL_TRIANGLES:
                    valid = get3(i, indices);
                    done = indices.remaining() < 3;
                    break;
                case GL11.GL_TRIANGLE_FAN:
                    int index = indices.get();
                    if(index == -1) {
                        valid = get3(i, indices);
                    } else {
                        i[1] = i[2];
                        i[2] = index;
                    }
                    done = indices.remaining() == 0;      
                    break;
                default:
                    Logger.INSTANCE.warning("Unsupported topology.");
                    return false;
            }
            if(valid && 
                    getPosition(p0, i[0], vertices, posOffset, stride) &&
                    getPosition(p1, i[1], vertices, posOffset, stride) &&
                    getPosition(p2, i[2], vertices, posOffset, stride)) {
                Vector3f.sub(p1, p0, p01);
                Vector3f.sub(p2, p0, p02);
                Vector3f.cross(p02, p01, n);
                addNormal(n, i[0], vertices, normalOffset, stride);
                addNormal(n, i[1], vertices, normalOffset, stride);
                addNormal(n, i[2], vertices, normalOffset, stride);
                touched.add(i[0]);
                touched.add(i[1]);
                touched.add(i[2]);
            }
        }
        for(int index : touched) {
            vertices.position(index * stride + normalOffset);
            n.load(vertices);
            n.normalise();
            vertices.position(index * stride + normalOffset);
            n.store(vertices);
        }
        return true;
    }
    
    private static boolean getPosition(Vector3f pos, int i, FloatBuffer vertices, int posOffset, int stride) {
        if(i * stride + posOffset < vertices.capacity()) {
            vertices.position(i * stride + posOffset);
            pos.load(vertices);
            return true;
        } else {
            return false;
        }
    }
    
    private static boolean addNormal(Vector3f normal, int i, FloatBuffer vertices, int normalOffset, int stride) {
        if(i * stride + normalOffset < vertices.capacity()) {
            vertices.position(i * stride + normalOffset);
            vertices.put(vertices.get(vertices.position()) + normal.x);
            vertices.put(vertices.get(vertices.position()) + normal.y);
            vertices.put(vertices.get(vertices.position()) + normal.z);
            return true;
        } else {
            return false;
        }
    }
    
    private static boolean get3(int target[], IntBuffer indices) {
        int current = 0;
        while(current != 3 && indices.remaining() != 0) {
            int index = indices.get();
            if(index == -1) {
                current = 0;
            } else {
                target[current++] = index;
            }
        }
        return current == 3;
    }
}
