/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package graphics.geometry;

import java.nio.FloatBuffer;
import java.util.LinkedList;
import java.util.List;
import logging.Logger;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL15;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.GL30;

/**
 *
 * @author nico3000
 */
public class VertexBuffer {
    /**
     * GLSL: VertexPos.
     */
    public static final int ATTR_POS = 0;
    /**
     * GLSL: VertexNormal.
     */
    public static final int ATTR_NORMAL = 1;
    /**
     * GLSL: VertexColor.
     */
    public static final int ATTR_TEX = 2;
    
    /**
     * Die ID des Buffers.
     */
    private int id = 0;
    /**
     * Die Laenge in Byte eines zusammenhaengenden Vertices.
     */
    private int stride = 0;
    /**
     * Liste der Vertex Attribute.
     */
    private final List<VertexAttribPointer> attributes = new LinkedList<>();
    
    /**
     * Bindet den Buffer.
     */
    public void bind() {
        GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, this.id);
    }
    
    /**
     * Initialisiert den Buffer.
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean init() {
        this.id = GL15.glGenBuffers();
        return this.id != 0;
    }
    
    /**
     * Zerstoert den Buffer und gibt die Resourcen frei.
     */
    public void destroy() {
        if(this.id != 0) {
            GL15.glDeleteBuffers(this.id);
            this.id = 0;
        }
    }
    
    /**
     * Transferiert Daten an den Buffer.
     * @param data Die Daten
     */
    public void setData(FloatBuffer data) {
        GL30.glBindVertexArray(0);
        this.bind();
        GL15.glBufferData(GL15.GL_ARRAY_BUFFER, data, GL15.GL_STATIC_DRAW);
    }
    
    /**
     * Bindet alle Attribute des Buffers, die mit <tt>addAttribute()</tt>
     * hinzugefuegt wurden.
     * @return <tt>false</tt>, falls der Buffer keine Attribute hat
     */
    public boolean bindAttributes() {
        if(this.attributes.isEmpty()) {
            return false;
        } else {
            for(VertexAttribPointer attr : this.attributes) {
                attr.bind();
            }
            return true;
        }
    }
    
    /**
     * Fuegt dem Buffer ein Vertexattribut hinzu.
     * @param index Der Index des Attributs
     * @param size Die Anzahl der Komponenten
     */
    public void addAttribute(int index, int size) {
        VertexAttribPointer attribute = new VertexAttribPointer();
        attribute.index = index;
        attribute.size = size;
        attribute.type = GL11.GL_FLOAT;
        attribute.offset = this.stride;
        this.stride += 4 * size;
        this.attributes.add(attribute);
    }
    
    
    
    /**
     * Speichert Informationen ueber ein Vertexattribut.
     */
    private class VertexAttribPointer {
        /**
         * Der Index des Attributs.
         */
        private int index;
        /**
         * Die Anzahl der Komponenten.
         */
        private int size;
        /**
         * Der Typ des Attributs.
         */
        private int type;
        /**
         * Der Offset in Byte des ersten Attributs im Buffer.
         */
        private int offset;
        
        /**
         * Aktiviert und bindet das Attribut.
         */
        private void bind() {
            GL20.glEnableVertexAttribArray(this.index);
            GL20.glVertexAttribPointer(this.index, this.size, this.type, false, VertexBuffer.this.stride, this.offset);
        }
    }  
}
