/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package graphics.shader;

import graphics.geometry.VertexBuffer;
import java.util.HashMap;
import java.util.Map;
import logging.Logger;
import gameapp.GameApp;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.GL31;
import util.Pair;

/**
 * Ein Effekt ist eine Kombination aus einem oder mehreren Shadern
 * unterschiedlichen Typs.
 * @author nico3000
 */
public class Effect {
    /**
     * Der aktuell benutzte Effekt.
     */
    private static Effect currentEffect = null;

    /**
     * Liefert den aktuell benutzten Effekt.
     * @return Der aktuell benutzte Effekt
     */
    public static Effect getCurrentEffect() {
        return currentEffect;
    }
    
    /**
     * Die ID des Shaderobjekts (Rueckgabe von <tt>glCreateProgram()</tt>).
     */
    private int id = 0;
    /**
     * Der Vertexshader des Effekts.
     */
    private VertexShader vs = new VertexShader();
    /**
     * Der Fragmentshader des Effekts.
     */
    private FragmentShader fs = new FragmentShader();
    /**
     * Map zum schnellen Zugriff auf Uniformbloecke.
     */
    private final Map<String, Pair<Integer, Integer>> uniformBlockIndices = new HashMap<>();
    
    /**
     * Initialisiert einen Effekt mit Vertex- und Fragmentshader.
     * @param vsResource Die Resource, die den Sourcecode des Vertexshaders
     *                   enthaelt
     * @param fsResource Die Resource, die den Sourcecode des Fragmentshaders
     *                   enthaelt
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean initVF(String vsResource, String fsResource) {
        this.id = GL20.glCreateProgram();
        if(this.id != 0 && this.vs.init(vsResource) && this.fs.init(fsResource)) {
            GL20.glAttachShader(this.id, this.vs.getId());
            GL20.glAttachShader(this.id, this.fs.getId());
            GL20.glBindAttribLocation(this.id, VertexBuffer.ATTR_POS, "VertexPos");
            GL20.glBindAttribLocation(this.id, VertexBuffer.ATTR_NORMAL, "VertexNormal");
            GL20.glBindAttribLocation(this.id, VertexBuffer.ATTR_TEX, "VertexTex");
            GL20.glLinkProgram(this.id);
            //System.out.println(GL20.glGetAttribLocation(this.id, "VertexPos"));
            //System.out.println(GL20.glGetAttribLocation(this.id, "VertexNormal"));
            //System.out.println(GL20.glGetAttribLocation(this.id, "VertexColor"));
            if(GL20.glGetProgrami(this.id, GL20.GL_LINK_STATUS) == GL11.GL_FALSE) {
                String log = GL20.glGetProgramInfoLog(this.id, 1024);
                Logger.INSTANCE.error(log);
            } else {
                this.bindUniformBlockToIndex("CameraBlock", UniformBuffer.BINDING_CAMERA);
                this.bindUniformBlockToIndex("ModelBlock", UniformBuffer.BINDING_MODEL);
                this.bindUniformBlockToIndex("MaterialBlock", UniformBuffer.BINDING_MATERIAL);
                return true;
            }
        }
        this.destroy();
        return false;        
    }
    
    /**
     * Validiert den Effekt gegen den aktuellen Status des Graphics Device.
     * @return <tt>true</tt>, wenn die Validierung keine Probleme erkannt hat
     */
    public boolean validate() {
        GL20.glValidateProgram(this.id);
        if(GL20.glGetProgrami(this.id, GL20.GL_VALIDATE_STATUS) == GL11.GL_FALSE) {
            int infoLength = GL20.glGetProgrami(this.id, GL20.GL_INFO_LOG_LENGTH);
            String info = GL20.glGetProgramInfoLog(this.id, infoLength);
            Logger.INSTANCE.error(info);
            return false;
        } else {
            return true;
        }
    }
    
    /**
     * Aktiviert diesen Effekt.
     */
    public void use() {
        if(currentEffect != this) {
            currentEffect = this;
            if(GameApp.isInDebugMode()) {
                this.validate();
            }
            GL20.glUseProgram(this.id);
        }
    }
    
    /**
     * Loescht den Effekt und alle seine Shader.
     */
    public void destroy() {
        this.vs.destroy();
        this.fs.destroy();
        if(this.id != 0) {
            GL20.glDeleteProgram(this.id);
            this.id = 0;
        }
    }
    
    /**
     * Bindet eine UniformBlock Variable dieses Shaders an einen bestimmten
     * Index von <tt>GL_UNIFORM_BUFFER</tt>. Diese Methode <i>muss</i> einmal
     * vor <tt>bindUniformBuffer()</tt> fuer <tt>blockName</tt> aufgerufen
     * werden.
     * @param blockName Der Name des Blocks im Shader
     * @param index Der Index von <tt>GL_UNIFORM_BUFFER</tt>
     */
    private void bindUniformBlockToIndex(String blockName, int bufferIndex) {
        Pair<Integer, Integer> pair = this.uniformBlockIndices.get(blockName);
        if(pair == null) {
            pair = new Pair<>(GL31.glGetUniformBlockIndex(this.id, blockName), bufferIndex);
            this.uniformBlockIndices.put(blockName, pair);
        }
        if(pair.left != GL31.GL_INVALID_INDEX) {
            GL31.glUniformBlockBinding(this.id, pair.left, pair.right);
        }
    }
    
    /**
     * Setzt eine Uniform Variable auf einen bestimmten Wert.
     * @param var Die Variable
     * @param val Der Wert
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean setUniform(String var, int val) {
        int loc = GL20.glGetUniformLocation(this.id, var);
        if(loc != -1) {
            GL20.glUniform1i(loc, val);
            return true;
        } else {
            //Logger.INSTANCE.info("Uniform " + var + " does not exist in shader or is unused.");
            return false;
        }
    }
}
