/*
 * This software is the proprietary information of nico3000.
 */
package gui;

import gameapp.GameApp;
import graphics.texture.Texture;
import graphics.texture.TextureLoader;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.font.GlyphVector;
import java.awt.font.LineBreakMeasurer;
import java.awt.font.TextAttribute;
import java.awt.font.TextLayout;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import java.text.AttributedCharacterIterator;
import java.text.AttributedString;
import java.util.LinkedList;
import java.util.List;

import logging.Logger;

import org.lwjgl.BufferUtils;

import resource.ResHandle;
import resource.Resource;
import resource.loader.ImageExtraData;

/**
 *
 * @author nico3000
 */
public class DefaultComponentDrawer implements ComponentDrawer {
    protected Color DEFAULT_COLOR_BACKGROUND = new Color(0.0f, 0.0f, 0.0f, 0.8f);
    protected Color HOVERED_COLOR_BACKGROUND = new Color(0.5f, 0.5f, 0.5f, 1.0f);
    protected Color DEFAULT_COLOR_BORDER = new Color(1.0f, 1.0f, 1.0f, 1.0f);
    protected Color HOVERED_COLOR_BORDER = new Color(1.0f, 1.0f, 1.0f, 1.0f);
    protected Color DEFAULT_COLOR_FONT = new Color(1.0f, 1.0f, 1.0f, 1.0f);
    protected Color WINDOW_COLOR_BACKGROUND = new Color(1.0f, 1.0f, 1.0f, 1.0f);
    protected Color WINDOW_COLOR_BORDER = new Color(0.0f, 0.0f, 0.0f, 1.0f);
    protected Color WINDOW_COLOR_TITLE = new Color(0.0f, 0.0f, 0.0f, 1.0f);
    protected Color WINDOW_COLOR_FONT = new Color(1.0f, 1.0f, 1.0f, 1.0f);
    protected Font DEFAULT_FONT_TITLE = new Font("Arial", Font.PLAIN, 16);
    protected Font DEFAULT_FONT_TEXT = new Font("Arial", Font.PLAIN, 16);
    protected static final int BORDER_LR = 4;
    protected static final int BORDER_TB = 4;    
    protected static final int ALIGN_V_T = 1;  // 0001
    protected static final int ALIGN_V_C = 3;  // 0011
    protected static final int ALIGN_V_B = 2;  // 0010
    protected static final int ALIGN_V = 3;    // 0011
    protected static final int ALIGN_H_L = 4;  // 0100
    protected static final int ALIGN_H_C = 12; // 1100
    protected static final int ALIGN_H_R = 8;  // 1000
    protected static final int ALIGN_H = 12;   // 1100
    protected int buttonPadding = 4;
    protected int labelPadding = 4;
    protected int cornerRadius = 12;
    
    
    public BufferedImage createBaseImage(int width, int height) {
        BufferedImage image = new BufferedImage(width, height, BufferedImage.TYPE_4BYTE_ABGR);
        Graphics2D graphics = image.createGraphics();
        graphics.setBackground(new Color(1.0f, 1.0f, 1.0f, 0.0f));
        graphics.clearRect(0, 0, image.getWidth(), image.getHeight());
        return image;
    } 

    @Override
    public boolean init() {
        DEFAULT_COLOR_BACKGROUND = GameApp.getConfig().getAsColor("cDefaultColorBackground", DEFAULT_COLOR_BACKGROUND);
        HOVERED_COLOR_BACKGROUND = GameApp.getConfig().getAsColor("cHoveredColorBackground", HOVERED_COLOR_BACKGROUND);
        DEFAULT_COLOR_BORDER = GameApp.getConfig().getAsColor("cDefaultColorBorder", DEFAULT_COLOR_BORDER);
        HOVERED_COLOR_BORDER = GameApp.getConfig().getAsColor("cHoveredColorBorder", HOVERED_COLOR_BORDER);
        DEFAULT_COLOR_FONT = GameApp.getConfig().getAsColor("cDefaultColorFont", DEFAULT_COLOR_FONT);
        WINDOW_COLOR_TITLE = GameApp.getConfig().getAsColor("cWindowColorTitle", WINDOW_COLOR_TITLE);
        WINDOW_COLOR_BACKGROUND = GameApp.getConfig().getAsColor("cWindowColorBackground", WINDOW_COLOR_BACKGROUND);
        WINDOW_COLOR_BORDER = GameApp.getConfig().getAsColor("cWindowColorBorder", WINDOW_COLOR_BORDER);
        WINDOW_COLOR_FONT = GameApp.getConfig().getAsColor("cWindowColorFont", WINDOW_COLOR_FONT);
        DEFAULT_FONT_TITLE = this.getFromString(GameApp.getConfig().getAsString("sDefaultFontTitle"), DEFAULT_FONT_TITLE);
        DEFAULT_FONT_TEXT = this.getFromString(GameApp.getConfig().getAsString("sDefaultFontText"), DEFAULT_FONT_TEXT);
        return true;
    }
    
    private Font getFromString(String s, Font defaultFont) {
        if(s == null) {
            return defaultFont;
        }
        String split[] = s.split(",");
        try {
            switch (split.length) {
                case 1:
                    return new Font(split[0].trim(), Font.PLAIN, 16);
                case 2:
                    return new Font(split[0], Font.PLAIN, Integer.parseInt(split[1].trim()));
                default:
                    return defaultFont;
            }
        } catch (NumberFormatException ex) {
            return defaultFont;
        }
    }

    @Override
    public Texture createLabel(String text, int width, int height, boolean decorated) {
        BufferedImage image = this.createBaseImage(width, height);
        if(decorated) {
            this.drawFrame(image, null, this.cornerRadius);
        }
        if(text != null) {
            Rectangle contentRect = new Rectangle(this.labelPadding, this.labelPadding, width - 2 * this.labelPadding, height - 2 * this.labelPadding);
            if(text.startsWith("res:")) {
                this.drawImage(text.substring(4), image, contentRect);
            } else {
                this.drawText(text, image, contentRect, DEFAULT_FONT_TEXT, DEFAULT_COLOR_FONT, ALIGN_H_C | ALIGN_V_C, 1);
            }
        }
        return TextureLoader.createTexture(image);
    }
    
    @Override
    public Texture createButton(String content, int width, int height, boolean focus, boolean hovered, boolean pressed) {
        BufferedImage image = this.createBaseImage(width, height);
        this.drawButton(content, image, null, pressed);
        return TextureLoader.createTexture(image);
    }
    
    protected void drawImage(String imageResource, BufferedImage target, Rectangle rect) {
        ResHandle handle = GameApp.getResCache().getHandle(new Resource(imageResource));
        if(handle == null) {
            Logger.INSTANCE.error(imageResource + " is not known!");
        } else {
            if(handle.getExtra() == null || !(handle.getExtra() instanceof ImageExtraData)) {
                Logger.INSTANCE.error(imageResource + " is no readable image!");
            }
            ImageExtraData extra = (ImageExtraData)handle.getExtra();

            FloatBuffer buffer = BufferUtils.createFloatBuffer(handle.getSize() / 4);
            ByteBuffer srcBuffer = ByteBuffer.wrap(handle.getBuffer());
            int max = extra.getWidth() * extra.getHeight() * extra.getChannels();
            for(int i=0; i < max; ++i) {
                buffer.put(srcBuffer.getFloat());
            }
            buffer.position(0);
            handle.unlock();
            BufferedImage image = new BufferedImage(extra.getWidth(), extra.getHeight(), BufferedImage.TYPE_4BYTE_ABGR);
            for(int y=0; y < extra.getHeight(); ++y) {
                for(int x=0; x < extra.getWidth(); ++x) {
                    float color[] = { 1.0f, 1.0f, 1.0f, 1.0f, };
                    for(int c=0; c < extra.getChannels(); ++c) {
                        color[c] = buffer.get();
                    }
                    image.setRGB(x, y, new Color(color[0], color[1], color[2], color[3]).getRGB());
                }
            }
            target.createGraphics().drawImage(image, rect.x, target.getHeight() - rect.height - rect.y, rect.width, rect.height, null);
        }
    }

    @Override
    public Texture createTextField(String text, int caretPosition, int fontSize, int width, int height) {
        BufferedImage image = this.createBaseImage(width, height);
        this.drawFrame(image, null, this.cornerRadius);
        image.createGraphics();
        Graphics2D graphics = image.createGraphics();
        graphics.setColor(DEFAULT_COLOR_FONT);
        Font font = DEFAULT_FONT_TITLE.deriveFont((float)fontSize);
        graphics.setFont(font);
        GlyphVector vec = font.createGlyphVector(graphics.getFontRenderContext(), text);
        Rectangle2D rect = vec.getVisualBounds();
        float x = (float)BORDER_LR;
        float y = (float)(-rect.getY() + 0.5 * ((double)image.getHeight() - rect.getHeight()));
        graphics.drawGlyphVector(vec, x, y);
        graphics.setColor(Color.BLACK);
        if(caretPosition != -1) {
            if(caretPosition < vec.getNumGlyphs()) {
                Rectangle bounds = vec.getGlyphPixelBounds(caretPosition, graphics.getFontRenderContext(), x, y);
                graphics.drawLine(bounds.x - 1, (int)y + 2, bounds.x - 1, (int)(y - rect.getHeight()) - 2);
            } else if(vec.getNumGlyphs() != 0) {
                Rectangle bounds = vec.getGlyphPixelBounds(caretPosition - 1, graphics.getFontRenderContext(), x, y);
                graphics.drawLine(bounds.x + bounds.width + 1, (int)y + 2, bounds.x + bounds.width + 1, (int)(y - rect.getHeight()) - 2);
            } else {
                graphics.drawLine(BORDER_LR, BORDER_TB + 2, BORDER_LR, BORDER_TB + fontSize + 2);
            }
        }
        return TextureLoader.createTexture(image);
    }

    @Override
    public Texture createWindow(String title, Rectangle titleRect, Rectangle closeButtonRect, boolean focused, boolean closeable, boolean closeButtonPressed, int width, int height) {
        int titleBarHeight = 24;
        BufferedImage image = this.createBaseImage(width, height);
        Graphics2D graphics = image.createGraphics();
        // background
        graphics.setColor(WINDOW_COLOR_BACKGROUND);
        graphics.fillRect(0, titleBarHeight + 4, width, height);
        // border
        graphics.setStroke(new BasicStroke(2.0f));
        graphics.setColor(WINDOW_COLOR_BORDER);
        graphics.drawRect(1, 1, width - 2, height- 2);
        graphics.drawRect(1, 1, width - 2, titleBarHeight + 2);
        // title bar
        graphics.setColor(focused ? WINDOW_COLOR_TITLE : WINDOW_COLOR_TITLE.darker());
        graphics.fillRect(2, 2, width - 4, titleBarHeight);
        // title bar text
        Rectangle titleTextRect = new Rectangle(titleRect.x + this.labelPadding, titleRect.y + this.labelPadding, titleRect.width - 2 * this.labelPadding, titleRect.height - 2 * this.labelPadding);
        this.drawText(title, image, titleTextRect, DEFAULT_FONT_TITLE, WINDOW_COLOR_FONT, ALIGN_H_L | ALIGN_V_C, 1);
        // close button
        this.drawButton(closeable ? "res:Textures/Close.png" : "res:Textures/CloseDisabled.png", image, closeButtonRect, closeButtonPressed);
        
        return TextureLoader.createTexture(image);
    }

    @Override
    public Texture drawTextBlock(String text, int width) {
        BufferedImage image = new BufferedImage(1, 1, BufferedImage.TYPE_4BYTE_ABGR);
        Graphics2D g = image.createGraphics();
        g.setFont(DEFAULT_FONT_TEXT);
        
        AttributedString attrString = new AttributedString(text);
        attrString.addAttribute(TextAttribute.FONT, DEFAULT_FONT_TEXT);
        AttributedCharacterIterator charIter = attrString.getIterator();
        LineBreakMeasurer measurer = new LineBreakMeasurer(charIter, g.getFontRenderContext());
        List<TextLayout> lines = new LinkedList<>();
        while(measurer.getPosition() < charIter.getEndIndex()) {
            int newLine = text.indexOf('\n', measurer.getPosition() + 1);
            if(newLine != -1) {
                lines.add(measurer.nextLayout(width - 8, newLine, false));
            } else {
                lines.add(measurer.nextLayout(width - 8));
            }
        }
        int height = 8 + lines.size() * g.getFontMetrics().getHeight();
        if(height == 0) {
            return TextureLoader.getErrorTexture2d();
        } else {
            image = this.createBaseImage(width, height);
            g = image.createGraphics();
            g.setFont(DEFAULT_FONT_TEXT);
            Color fontColor = DEFAULT_COLOR_FONT;
            Color shadowColor = new Color(
                    255 - fontColor.getRed(),
                    255 - fontColor.getGreen(),
                    255 - fontColor.getBlue(),
                    fontColor.getAlpha());
            int y = 4;
            for(TextLayout layout : lines) {
                g.setColor(shadowColor);
                layout.draw(g, 5, y + layout.getAscent() + 1);
                g.setColor(fontColor);
                layout.draw(g, 4, y + layout.getAscent()); 
                y += g.getFontMetrics().getHeight();
            }
            return TextureLoader.createTexture(image);
        }
    }

    @Override
    public Texture createScrollbar(int width, int height, float scrolled) {
        BufferedImage image = this.createBaseImage(width, height);
        Graphics2D graphics = image.createGraphics();
        graphics.setColor(Color.BLUE);
        graphics.fillRect(0, 0, width, width);
        graphics.fillRect(0, height - width, width, width);
        graphics.setColor(Color.GREEN);
        graphics.fillRect(0, width + (int)((float)(height - 2 * width) * scrolled), width, width);
        return TextureLoader.createTexture(image);
    }
    
    private void drawButton(String content, BufferedImage target, Rectangle rect, boolean pressed) {
        if(rect == null) {
            rect = new Rectangle(0, 0, target.getWidth(), target.getHeight());
        }
        Graphics2D graphics = target.createGraphics();
        graphics.setColor(DEFAULT_COLOR_BACKGROUND);
        graphics.fill3DRect(rect.x, target.getHeight() - rect.height - rect.y, rect.width, rect.height, !pressed);
        Rectangle contentRect = new Rectangle(rect.x + this.buttonPadding, rect.y + this.buttonPadding, rect.width - 2 * this.buttonPadding, rect.height - 2 * this.buttonPadding);
        if(content != null) {
            if(content.startsWith("res:")) {
                this.drawImage(content.substring(4), target, contentRect);
            } else {
                this.drawText(content, target, contentRect, DEFAULT_FONT_TEXT, DEFAULT_COLOR_FONT, ALIGN_H_C | ALIGN_V_C, pressed ? 2 : 1);
            }
        }
    }
    
    protected void drawText(String text, BufferedImage target, Rectangle rect, Font font, Color color, int alignment, int shadowOffset) {
        double x = 0.0;
        double y = 0.0;
        Graphics2D graphics = target.createGraphics();
        graphics.setFont(font);
        GlyphVector vector = font.createGlyphVector(graphics.getFontRenderContext(), text);
        switch(alignment & ALIGN_H) {
            case ALIGN_H_C:
                x = 0.5 * ((double)rect.getWidth() - vector.getVisualBounds().getWidth());
                break;
            case ALIGN_H_R:
                x = (double)rect.getWidth() - vector.getVisualBounds().getWidth();
                break;
            case ALIGN_H_L:
                // nothing to do
                break;
        }
        switch(alignment & ALIGN_V) {
            case ALIGN_V_C:
                y = 0.5 * (vector.getVisualBounds().getHeight() + rect.getHeight());
                break;
            case ALIGN_V_T:
                y = vector.getVisualBounds().getHeight();
                break;
            case ALIGN_V_B:
                y = rect.getHeight();
                break;
        }        
        graphics.setColor(getInverse(color));
        graphics.drawString(text, (float)(rect.x + x) + (float)shadowOffset, (float)(target.getHeight() - rect.height - rect.y + y) + (float)shadowOffset);
        graphics.setColor(color);
        graphics.drawString(text, (float)(rect.x + x), (float)(target.getHeight() - rect.height - rect.y + y));
    }
    
    public void drawFrame(BufferedImage image, Rectangle rect, int cornerRadius) {
        if(rect == null) {
            rect = new Rectangle(0, 0, image.getWidth(), image.getHeight());
        }
        Graphics2D graphics = image.createGraphics();
        graphics.setColor(DEFAULT_COLOR_BACKGROUND);
        graphics.fillRoundRect(0, 0, rect.width - 1, rect.height - 1, cornerRadius, cornerRadius);
        graphics.setColor(DEFAULT_COLOR_BORDER);
        graphics.drawRoundRect(0, 0, rect.width - 1, rect.height - 1, cornerRadius, cornerRadius);
    }
    
    private static Color getInverse(Color src) {
        return new Color(255 - src.getRed(), 255 - src.getGreen(), 255 - src.getBlue(), src.getAlpha());
    }

}
