/*
 * Copyright (C) 2013 nico3000 - All Rights Reserved
 */
package mesh.loader.wavefront;

import graphics.geometry.Geometry;
import graphics.geometry.IndexBuffer;
import graphics.geometry.VertexBuffer;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import logging.Logger;
import gameapp.GameApp;
import mesh.IndexBufferInterval;
import mesh.loader.GeometryLoader;
import mesh.material.Material;
import mesh.material.MaterialLibrary;
import org.lwjgl.BufferUtils;
import org.lwjgl.util.vector.Vector2f;
import org.lwjgl.util.vector.Vector3f;
import org.w3c.dom.Node;
import resource.ResHandle;
import resource.Resource;
import util.XMLUtil;

/**
 * Erzeugt aus den Daten einer Wavefront <tt>.obj</tt> Datei eine Geometry.
 * @author nico3000
 */
public class WavefrontObjLoader implements GeometryLoader {
    /**
     * Name der Resource.
     */
    private final String resource;
    /**
     * Vertices der Datei.
     */
    private final List<Vector3f> vertexList = new ArrayList<>();
    /**
     * Normalen der Datei.
     */
    private final List<Vector3f> normalList = new ArrayList<>();
    /**
     * Texturkoordinaten der Datei.
     */
    private final List<Vector2f> texList = new ArrayList<>();
    /**
     * Faces der Datei.
     */
    private final Map<Material, List<Face>> faces = new HashMap<>();
    /**
     * Enthaelt waehrend des Ladens die aktuelle Zeile.
     */
    private int currentLine;
    /**
     * Enthaelt waehrend des Ladens die aktuelle Gruppe.
     */
    private String currentGroup = null;
    /**
     * Die aktuelle Materialbibliothek.
     */
    private MaterialLibrary currentMaterialLibrary = null;
    /**
     * Die vom Benutzer angegebene MaterialLibrary.
     */
    private MaterialLibrary customMaterialLibrary = null;
    /**
     * Das aktuelle Material.
     */
    private Material currentMaterial = null;
    /**
     * Alle referenzierten Materialbibliotheken.
     */
    private final Set<MaterialLibrary> materialLibraries = new HashSet<>();
    /**
     * Die Geometry.
     */
    private Geometry geometry;
    /**
     * Gibt an, ob das Model an der x-Achse gespiegelt werden soll.
     */
    private boolean mirrored = false;
    
    /**
     * Konstruktor.
     * @param resource Der Name der Resource
     */
    public WavefrontObjLoader(String resource) {
        this.resource = resource;
    }
    
    @Override
    public Geometry load() {
        this.geometry = Geometry.createGeometry(this.resource);
        if(this.geometry.isInitialized()) {
            return this.geometry;
        }
        ResHandle handle = GameApp.getResCache().getHandle(new Resource(this.resource));
        if(handle != null) {
            boolean success = false;
            BufferedReader reader = new BufferedReader(new InputStreamReader(handle.getBufferAsInputStream()));
            try {
                success = this.loadData(reader);
                reader.close();
            } catch (IOException ex) {
                Logger.INSTANCE.error(ex.getMessage());
            }
            handle.unlock();
            if(!success || !this.buildGeometry()) {
                this.geometry.release();
                this.geometry = null;
            }
        }
        return this.geometry;
    }
    
    /**
     * Extrahiert die Daten aus einem BufferedReader.
     * @param reader Der Inhalt der Datei
     * @return <tt>true</tt>, falls erfolgreich
     * @throws IOException Fehler des BufferedReader werden weitergereicht
     */
    private boolean loadData(BufferedReader reader) throws IOException {
        this.vertexList.clear();
        this.texList.clear();
        this.normalList.clear();
        this.faces.clear();
        String line;
        this.currentLine = 0;
        while((line = reader.readLine()) != null) {
            ++this.currentLine;
            line = line.trim();
            if(!line.isEmpty() && !line.startsWith("#")) {
                StringTokenizer tokenizer = new StringTokenizer(line.trim());
                String type = tokenizer.nextToken();
                boolean success;
                switch(type) {
                    case "v":
                        success = this.addVector3f(tokenizer, this.vertexList);
                        break;
                    case "vn":
                        success = this.addVector3f(tokenizer, this.normalList);
                        break;
                    case "vt":
                        success = this.addVector2f(tokenizer, this.texList);
                        break;
                    case "f":
                        success = this.addFace(tokenizer);
                        break;
                    case "g":
                        success = this.setGroup(tokenizer);
                        break;
                    case "usemtl":
                        success = this.setMaterial(tokenizer);
                        break;
                    case "mtllib":
                        String split[] = line.split(" ", 2);
                        if(split.length != 2) {
                            this.error("illegal mtllib definition");
                            success = false;
                        } else {
                            success = this.setMtlLib(split[1]);
                        }
                        break;
                    default:
                        //Logger.INSTANCE.warning(String.format("unsupported or illegal type \"%s\" was ignored (%s, %d)", type, this.resource, this.currentLine));
                        success = true;
                        break;
                }
                if(!success) {
                    return false;
                }
            }
        }
        if(this.currentMaterial != null) {
            this.currentMaterial.release();
        }
        int intervalStart = 0;
        for(Material material : this.faces.keySet()) {
            int indexCount = 0;
            for(Face face : this.faces.get(material)) {
                indexCount += face.getIndexCount() + 1;
            }
            this.geometry.assignIntervalToMaterial(new IndexBufferInterval(intervalStart, indexCount), material.getLibraryName(), material.getName());
            intervalStart += indexCount;
        }
        return true;
    }

    /**
     * Setzt die Materialbibliothek.
     * @param data Der Dateiname der Bibliothek
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean setMtlLib(String library) {
    	WavefrontMtlLoader loader = new WavefrontMtlLoader(library);
        this.currentMaterialLibrary = loader.load();
        if(this.currentMaterialLibrary != null) {
            if(!this.materialLibraries.add(this.currentMaterialLibrary)) {
                this.currentMaterialLibrary.release();
            }
            return true;
        } else {
            return false;
        }
    }
    
    /**
     * Setzt die aktuelle Gruppe.
     * @param data Der Name der Gruppe
     * @return <tt>true</tt>, falls erfolgreich
     */
    private boolean setGroup(StringTokenizer data) {
        if(data.countTokens() != 1) {
            this.currentGroup = null;
        } else {
            this.currentGroup = data.nextToken();
        }
        return true;
    }
    
    /**
     * Setzt das aktuelle Material.
     * @param data Der Name des Materials
     * @return <tt>true</tt>, falls erfolgreich
     */
    private boolean setMaterial(StringTokenizer data) {
        if(data.countTokens() != 1) {
            this.error("illegal material usage");
            return false;            
        }
        String name = data.nextToken();
        if(this.currentMaterial != null) {
            this.currentMaterial.release();
        }
        if(this.currentMaterialLibrary == null) {
            this.error("No current MaterialLibrary for Material %s!", name);
            return false;
        }
        this.currentMaterial = this.currentMaterialLibrary.getMaterialByName(name);
        if(this.currentMaterial == null) {
            this.error("Failed to load requested Material %s!", name);
            return false;
        } else {
            return true;
        }
    }
    
    /**
     * Wandelt die Daten eines Strings in einen Vector3f um und fuegt ihn bei
     * Erfolg einer Liste hinzu.
     * @param data Die Daten
     * @param list Die Liste
     * @return <tt>true</tt>, falls erfolgreich
     */
    private boolean addVector3f(StringTokenizer data, List<Vector3f> list) {
        if(data.countTokens() < 3) {
            this.error("wrong data count");
            return false;
        }
        float x, y, z;
        try {
            x = Float.parseFloat(data.nextToken());
            y = Float.parseFloat(data.nextToken());
            z = Float.parseFloat(data.nextToken());
        } catch (NumberFormatException ex) {
            this.error("illegal float expression");
            return false;
        }        
        list.add(new Vector3f(x, y, z));
        return true;
    }
    
    /**
     * Wandelt die Daten eines Strings in einen Vector2f um und fuegt ihn bei
     * Erfolg einer Liste hinzu.
     * @param data Die Daten
     * @param list Die Liste
     * @return <tt>true</tt>, falls erfolgreich
     */
    private boolean addVector2f(StringTokenizer data, List<Vector2f> list) {
        if(data.countTokens() < 2) {
            this.error("wrong data count");
            return false;
        }
        float x, y;
        try {
            x = Float.parseFloat(data.nextToken());
            y = Float.parseFloat(data.nextToken());
        } catch (NumberFormatException ex) {
            this.error("illegal float expression");
            return false;
        }
        list.add(new Vector2f(x, y));
        return true;
    }
    
    /**
     * Wandelt die Daten eines Strings in ein Face um, wie es die
     * obj-Spezifikation vorsieht.
     * @param data Die Daten
     * @return <tt>true</tt>, falls erfolgreich
     */
    private boolean addFace(StringTokenizer data) {
        int count = data.countTokens();
        if(count < 2) {
            this.error("Illegal face definition!");
            return false;
        }
        Face face = new Face(count);
        for(int i=0; i < count; ++i) {
            String split[] = data.nextToken().split("/");
            FaceType type = this.getFaceType(split);
            if(type == FaceType.UNKOWN) {
                this.error("Illegal face index %d!", i);
                return false;
            }
            if(face.type == FaceType.UNKOWN) {
                face.type = type;
            } else if(face.type != type) {
                this.error("Indexing type of face dffers from other faces!");
                return false;
            }
            try {
                face.setVertexIndex(i, Integer.parseInt(split[0]));
                switch (face.type) {
                    case VERTEX_TEXTURE_NORMAL:
                        face.setNormalIndex(i, Integer.parseInt(split[2]));     
                        // no break
                    case VERTEX_TEXTURE:
                        face.setTexIndex(i, Integer.parseInt(split[1]));
                        break;
                    case VERTEX_NORMAL:
                        face.setNormalIndex(i, Integer.parseInt(split[2]));
                        break;
                }
            } catch (NumberFormatException ex) {
                this.error("Illegal int expression!");
                return false;
            }
        }
        List<Face> list = this.faces.get(this.currentMaterial);
        if(list == null) {
            this.faces.put(this.currentMaterial, list = new LinkedList<>());
        }
        list.add(face);
        return true;
    }
    
    /**
     * Bestimmt aus einem an <tt>/</tt> geteilten Strings die Art des Faces.
     * @param split Der an <tt>t</tt> geteilte String
     * @return Der Typ des Faces
     */
    private FaceType getFaceType(String split[]) {
        if(split.length == 1) {
            return FaceType.VERTEX;
        } else if(split.length == 2) {
            return FaceType.VERTEX_TEXTURE;
        } else if(split.length == 3) {
            if(split[1].isEmpty()) {
                return FaceType.VERTEX_NORMAL;
            } else {
                return FaceType.VERTEX_TEXTURE_NORMAL;
            }
        }
        return FaceType.UNKOWN;
    }
    
    /**
     * Loggt einen Fehler und fuegt den Namen und die Zeile der obj-Datei an.
     * @param message Die Fehlermeldung
     */
    private void error(String message, Object... args) {
        String formatted = String.format(message, args);
        Logger.INSTANCE.error(1, "%s (%s, %d)", formatted, this.resource, this.currentLine);
    }
    
    private Vector3f getVertex(int i) {
        return this.vertexList.get(i);
    }
    
    private Vector3f getNormal(int i) {
        return i < this.normalList.size() ? this.normalList.get(i) : new Vector3f();
    }
    
    private Vector2f getTexture(int i) {
        return i < this.texList.size() ? this.texList.get(i) : new Vector2f();
    }

    @Override
    public void setAdditionalData(Node data) {
        if(data != null) {
            String customLibraryName = XMLUtil.getAttributeAsString(data, "customMaterialLibrary");
            if(customLibraryName != null) {
                WavefrontMtlLoader loader = new WavefrontMtlLoader(customLibraryName);
                this.customMaterialLibrary = loader.load();
                if(this.customMaterialLibrary != null) {
                    //Logger.INSTANCE.info("Using custom MaterialLibrary %s for Mesh %s.", customLibraryName, this.resource);
                } else {
                    Logger.INSTANCE.warning("Loading of custom MaterialLibrary %s for Mesh %s failed.", customLibraryName, this.resource);
                }
            }
        }
    }
    
    private boolean buildGeometry() {
        IndexBuffer indexBuffer = new IndexBuffer();
        if(!indexBuffer.init()) {
            return false;
        }
        VertexBuffer vertexBuffer = new VertexBuffer();
        if(!vertexBuffer.init()) {
            indexBuffer.destroy();
            return false;
        }
        vertexBuffer.addAttribute(VertexBuffer.ATTR_POS, 3);
        vertexBuffer.addAttribute(VertexBuffer.ATTR_NORMAL, 3);
        vertexBuffer.addAttribute(VertexBuffer.ATTR_TEX, 2);
        
        List<Float> vertices = new LinkedList<>();
        List<Integer> indices = new LinkedList<>();
        Map<WavefrontObjLoader.ObjIndex, Integer> indexMappings = new HashMap<>();
        boolean generateNormals = false;
        for(Material material : this.faces.keySet()) {
            for(Face face : this.faces.get(material)) {
                if(face.getType() == FaceType.VERTEX || face.getType() == FaceType.VERTEX_TEXTURE) {
                    generateNormals = true;
                }
                for(int i = 0; i < face.getIndexCount(); ++i) {
                    ObjIndex objIndex = face.getIndex(this.mirrored ? i : face.getIndexCount() - 1 - i);
                    Integer index = indexMappings.get(objIndex);
                    if(index == null) {
                        Vector3f vertex = this.getVertex(objIndex.getVertex());
                        Vector3f normal = this.getNormal(objIndex.getNormal());
                        Vector2f tex = this.getTexture(objIndex.getTex());
                        index = vertices.size() / 8;
                        vertices.add(this.mirrored ? -vertex.x : vertex.x);
                        vertices.add(vertex.y);
                        vertices.add(vertex.z);
                        vertices.add(this.mirrored ? -normal.x : normal.x);
                        vertices.add(normal.y);
                        vertices.add(normal.z);
                        vertices.add(tex.x);
                        vertices.add(tex.y);
                        indexMappings.put(objIndex, index);
                    }
                    indices.add(index);
                }
                indices.add(IndexBuffer.RESTART);
            }
        }
        IntBuffer indexIntBuffer = BufferUtils.createIntBuffer(indices.size());
        for(Integer index : indices) {
            indexIntBuffer.put(index);
        }
        indexIntBuffer.position(0);
        indexBuffer.setData(indexIntBuffer, IndexBuffer.TRIANGLE_FAN);
        FloatBuffer vertexFloatBuffer = BufferUtils.createFloatBuffer(vertices.size());
        for(Float f : vertices) {
            vertexFloatBuffer.put(f);
        }
        if(generateNormals && !Geometry.calculateNormals(vertexFloatBuffer, 0, 3, 8, indexIntBuffer, IndexBuffer.TRIANGLE_FAN)) {
            Logger.INSTANCE.warning("Normal calculation failed.");
        }
        vertexFloatBuffer.position(0);
        vertexBuffer.setData(vertexFloatBuffer);
        boolean success = this.geometry.init(indexBuffer, vertexBuffer);
//        if(this.resource.endsWith("cube.obj")) {
//            String vertexOut = "";
//            for(int i=0; i < vertexFloatBuffer.capacity(); i += 8) {
//                vertexOut += String.format("%d: Pos: %.2f, %.2f, %.2f\n", i / 8, vertexFloatBuffer.get(i), vertexFloatBuffer.get(i+1), vertexFloatBuffer.get(i+2));
//                //vertexOut += String.format("Normal: %.2f, %.2f, %.2f\n", vertices.get(i+3), vertices.get(i+4), vertices.get(i+5));
//                //vertexOut += String.format("Tex: %.2f, %.2f\n\n", vertices.get(i+6), vertices.get(i+7));
//            }
//            Logger.INSTANCE.info("Printing cube.obj detail\nVertices\n%s----", vertexOut);
//            
//            String out = "";
//            for(int i=0; i < indexIntBuffer.capacity(); ++i) {
//                int index = indexIntBuffer.get(i);
//                out += index;
//                if(index != IndexBuffer.RESTART) {
//                    out += " ";
//                } else {
//                    out += "\n";
//                }
//            }
//            Logger.INSTANCE.info("Indices\n%s----", out);
//        }
        if(!success) {
            vertexBuffer.destroy();
            indexBuffer.destroy();
            return false;
        } else {
            return true;
        }    
    }

    @Override
    public void release() {
        for(MaterialLibrary library : this.materialLibraries) {
            library.release();
        }
        this.materialLibraries.clear();
        if(this.customMaterialLibrary != null) {
            this.customMaterialLibrary.release();
            this.customMaterialLibrary = null;
        }
    }

    @Override
    public String getCustomMaterialLibrary() {
        return this.customMaterialLibrary == null ? null : this.customMaterialLibrary.getName();
    }
    
    /**
     * Moegliche Typen von Faces in einer obj-Datei.
     */
    public enum FaceType {
        /**
         * Unbekannt / Unegueltig.
         */
        UNKOWN,
        /**
         * Nur Vertices.
         */
        VERTEX,
        /**
         * Vertices und Texturkoordinaten.
         */
        VERTEX_TEXTURE,
        /**
         * Vertices und Normalen.
         */
        VERTEX_NORMAL,
        /**
         * Vertices, Texturkoordinaten und Normalen.
         */
        VERTEX_TEXTURE_NORMAL,
    }
    
    public class ObjIndex {
        private int vertex;
        private int normal;
        private int tex;

        private ObjIndex(int vertex, int normal, int tex) {
            this.vertex = vertex;
            this.normal = normal;
            this.tex = tex;
        }

        public int getVertex() {
            return vertex;
        }

        public int getNormal() {
            return normal;
        }

        public int getTex() {
            return tex;
        }

        @Override
        public int hashCode() {
            int hash = 3;
            hash = 61 * hash + this.vertex;
            hash = 61 * hash + this.normal;
            hash = 61 * hash + this.tex;
            return hash;
        }

        @Override
        public boolean equals(Object obj) {
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final ObjIndex other = (ObjIndex) obj;
            if (this.vertex != other.vertex) {
                return false;
            }
            if (this.normal != other.normal) {
                return false;
            }
            if (this.tex != other.tex) {
                return false;
            }
            return true;
        }
    }
    
    /**
     * Sammelt die Indizes eines Faces.
     */
    public class Face {
        /**
         * Die Indizes des Faces.
         */
        private final ObjIndex indices[];
        /**
         * Der Typ des Faces.
         */
        private FaceType type = FaceType.UNKOWN;
        /**
         * Gruppe des Faces.
         */
        private final String group;
        
        /**
         * Konstruktor.
         * @param indexCount Die Anzahl der Indizes
         */
        private Face(int indexCount) {
            this.indices = new ObjIndex[indexCount];
            this.group = WavefrontObjLoader.this.currentGroup;
            for(int i=0; i < indexCount; ++i) {
                this.indices[i] = new ObjIndex(0, 0, 0);
            }
        }
        
        /**
         * Fuegt den Index eines Vertices dem Face hinzu. Ist der Index negativ
         * (stellt also eine relative Indizierung dar), wird dieser in einen
         * absoluten Index ueberfuehrt.
         * @param pos Die Nummer des Indizes
         * @param vertexIndex Der Index
         */
        private void setVertexIndex(int pos, int vertexIndex) {
            if(vertexIndex < 0) {
                this.indices[pos].vertex = WavefrontObjLoader.this.vertexList.size() + vertexIndex;
            } else {
                this.indices[pos].vertex = vertexIndex - 1;
            }
        }
        
        /**
         * Fuegt den Index einer Texturkoordinate dem Face hinzu. Ist der Index
         * negativ (stellt also eine relative Indizierung dar), wird dieser in
         * einen absoluten Index ueberfuehrt.
         * @param pos Die Nummer des Indizes
         * @param texIndex Der Index
         */
        private void setTexIndex(int pos, int texIndex) {
            if(texIndex < 0) {
                this.indices[pos].tex = WavefrontObjLoader.this.texList.size() + texIndex;
            } else {
                this.indices[pos].tex = texIndex - 1;
            }
        }
        
        /**
         * Fuegt den Index einer Normalen dem Face hinzu. Ist der Index negativ
         * (stellt also eine relative Indizierung dar), wird dieser in einen
         * absoluten Index ueberfuehrt.
         * @param pos Die Nummer des Indizes
         * @param normalIndex Der Index
         */        
        private void setNormalIndex(int pos, int normalIndex) {
            if(normalIndex < 0) {
                this.indices[pos].normal = WavefrontObjLoader.this.normalList.size() + normalIndex;
            } else {
                this.indices[pos].normal = normalIndex - 1;
            }
        }

        public FaceType getType() {
            return this.type;
        }
        
        public int getIndexCount() {
            return this.indices.length;
        }
        
        public ObjIndex getIndex(int i) {
            return this.indices[i];
        }
    }
}
