/*
 * This software is the proprietary information of nico3000.
 */
package n3gui;

import logging.Logger;
import gameapp.GameApp;
import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Vector4f;
import util.MathUtil;

/**
 *
 * @author nico3000
 */
public class Dimension {
    /**
     * Position is relative to parent's left border.
     */
    public static final int REF_H_L = 1;        // ...0001
    /**
     * Position is relative to parent's right border.
     */
    public static final int REF_H_R = 2;        // ...0010
    /**
     * Position is relative to parent's horizontal center.
     */
    public static final int REF_H_C = 3;        // ...0011
    /**
     * Mask for horizontal reference point.
     */
    private static final int REF_H_MASK = 3;    // ...0011    
    /**
     * Position is relative to parent's top border.
     */
    public static final int REF_V_T = 4;        // ...0100
    /**
     * Position is relative to parent's bottom border.
     */
    public static final int REF_V_B = 8;        // ...1000
    /**
     * Position is relative to parent's vertical center.
     */
    public static final int REF_V_C = 12;       // ...1100
    /**
     * Mask for vertical reference point.
     */
    private static final int REF_V_MASK = 12;   // ...1100
    /**
     * Temporaere Skalierungsmatrix fuer die Methode <tt>getTransform()</tt>.
     */
    private static final Matrix4f TEMP_SCALE = new Matrix4f();
    /**
     * Temporaere Translationsmatrix fuer die Methode <tt>getTransform()</tt>.
     */
    private static final Matrix4f TEMP_TRANSLATION = new Matrix4f();
    
    /**
     * Der zugehoerige Knoten.
     */
    private final GuiTreeNode node;
    /**
     * Zeigt an, ob sich die Position seit dem letzten Aufruf von
     * <tt>getTransform()</tt> geaendert hat. Diese Methode wird in der Methode
     * <tt>draw()</tt> des <tt>GuiTreeNodes</tt> aufgerufen.
     */
    private boolean posChanged = true;
    /**
     * Zeigt an, ob sich die Position seit dem letzten Aufruf von
     * <tt>getTransform()</tt> geaendert hat. Diese Methode wird in der Methode
     * <tt>draw()</tt> des <tt>GuiTreeNodes</tt> aufgerufen.
     */
    private boolean sizeChanged = true;    
    /**
     * Die Transformationsmatrix der Dimension. Mit dieser Matrix wird ein um
     * den Nullpunkt zentriertes bildschirmfuellendes Rechteck an die von dieser
     * Dimension repraesentierten Stelle transformiert.
     */
    private final Matrix4f transform = Matrix4f.setIdentity(new Matrix4f());
    /**
     * Der Referenzpunkt, auf den sich die Positionsangaben beziehen.
     */
    private int referencePoint = REF_H_L | REF_V_T;
    /**
     * Die x-Koordinate des Referenzpunktes bezueglich des Referenzpunktes der
     * uebergeordneten Dimension.
     */
    private int posX = 0;
    /**
     * Die y-Koordinate des Referenzpunktes bezueglich des Referenzpunktes der
     * uebergeordneten Dimension.
     */
    private int posY = 0;
    /**
     * Die Breite der Dimension in Pixel.
     */
    private int width = 0;
    /**
     * Die Hoehe der Dimension in Pixel.
     */
    private int height = 0;
    /**
     * Der obere linke und untere rechte Punkt des Clippingrechtecks in
     * Bildchirmkoordinaten.
     */
    private final Vector4f clippingRect = new Vector4f(-1.0f, 1.0f, 1.0f, -1.0f);
    /**
     * Gibt an, ob sich Eigenschaften der Dimension geaendert haben.
     */
    private boolean changed = true;

    /**
     * Konstruktor.
     * @param node Der zugehoerige Knoten
     */
    public Dimension(GuiTreeNode node) {
        this.node = node;
    }
    
    /**
     * Setzt den Referenzpunkt, auf den sich die Positionsangaben beziehen. Der
     * Punkt wird mit einer Oder-Verknuepfung einer <tt>REF_H_*</tt> und einer
     * <tt>REF_V_*</tt> Konstanten beschrieben. Beispielsweise wird die linke
     * obere Ecke bei <tt>REF_H_L | REF_H_T</tt> als Referenzpunkt gewaehlt.
     * Dieser Punkt ist auch standardmaessig gewaehlt.
     * @param referencePoint Der Referenzpunkt
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean setReferencePoint(int referencePoint) {
        if((referencePoint & REF_H_MASK) == 0) {
            Logger.INSTANCE.warning("No horizontal reference point given.");
            return false;
        } else if((referencePoint & REF_V_MASK) == 0) {
            Logger.INSTANCE.warning("No vertical reference point given.");
            return false;
        } else if(this.referencePoint != referencePoint) {
            this.positionChanged();
            this.referencePoint = referencePoint;
            return true;
        } else {
            return true;
        }
    }
    
    /**
     * Bewegt die Dimension entlang eines Vektors.
     * @param dx Die x-Komponente des Vektors
     * @param dy Die y-Komponente des Vektors
     */
    public void move(int dx, int dy) {
        this.setPosition(this.posX + dx, this.posY + dy);
    }
    
    /**
     * Setzt die Position der Dimension bezueglich des Referenzpunktes.
     * @param posX Die x-Koordinate des Referenzpunktes
     * @param posY Die y-Koordinate des Referenzpunktes
     */
    public void setPosition(int posX, int posY) {
        if(this.posX != posX || this.posY != posY) {
            this.posX = posX;
            this.posY = posY;
            this.positionChanged();
        }
    }
    
    /**
     * Teilt der Dimension mit, dass sich das Elternelement geaendert hat.
     */
    public void parentChanged() {
        this.positionChanged();
    }
    
    /**
     * Teilt der Dimension mit, dass sich ihre Position veraendert hat.
     */
    void positionChanged() {
        this.posChanged = true;
        for(GuiTreeNode child : this.node.getChildren()) {
            child.getDimension().positionChanged();
        }
    }
    
    /**
     * Setzt die Groesse der Dimension in Pixel.
     * @param width Die Breite in Pixel
     * @param height Die Hoehe in Pixel
     */
    public void setSize(int width, int height) {
        if(this.width != width || this.height != height) {
            this.width = width;
            this.height = height;
            this.sizeChanged();
        }
    }
    
    /**
     * Teilt der Dimension mit, dass sich ihre Groesse geaendert hat.
     */
    void sizeChanged() {
        this.sizeChanged = true;
        if(this.referencePoint != (REF_H_L | REF_V_T)) {
            this.positionChanged();
        }
        for(GuiTreeNode child : this.node.getChildren()) {
            child.getDimension().parentsSizeChanged();
            this.getClippingRect(child.getDimension().clippingRect);
        }
    }
    
    /**
     * Wird aufgerufen, wenn die Vaterdimension die Groesse geaendert hat.
     */
    private void parentsSizeChanged() {
        if(this.referencePoint != (REF_H_L | REF_V_T)) {
            this.positionChanged();
        }
    }
    
    /**
     * Liefert die Breite der Dimension in Pixel.
     * @return Die Breite
     */
    public int getWidth() {
        return this.width;
    }
    
    /**
     * Liefert die Hoehe der Dimension in Pixel.
     * @return Die Hoehe
     */
    public int getHeight() {
        return this.height;
    }

    /**
     * Liefert, ob sich die Position dieser oder der uebergeordneten Dimension
     * geaendert hat.
     * @return <tt>true</tt>, falls die Position sich geaendert hat
     */
    public boolean hasPosChanged() {
        return this.posChanged;
    }

    /**
     * Liefert, ob sich die Groesse dieser Dimension geaendert hat.
     * @return <tt>true</tt>, falls die Groesse sich geaendert hat
     */
    public boolean hasSizeChanged() {
        return this.sizeChanged;
    }
    
    void clearChangedStates() {
        this.sizeChanged = false;
        this.posChanged = false;
    }

    /**
     * Liefert die Transformationsmatrix.
     * @return Die Matrix
     */
    public Matrix4f getTransform() {
        if(this.sizeChanged || this.posChanged) {
            float quadScaleX = 2.0f * (float)this.width / (float)GameApp.getGraphicsLayer().getWidth();
            float quadScaleY = -2.0f * (float)this.height / (float)GameApp.getGraphicsLayer().getHeight();
            float quadTransX = 2.0f * (float)this.getGlobalX() / (float)GameApp.getGraphicsLayer().getWidth() - 1.0f;
            float quadTransY = -2.0f * (float)this.getGlobalY() / (float)GameApp.getGraphicsLayer().getHeight() + 1.0f;
            MathUtil.scaling(TEMP_SCALE, quadScaleX, quadScaleY, 1.0f);
            MathUtil.translation(TEMP_TRANSLATION, quadTransX, quadTransY, 0.0f);
            MathUtil.mul(this.transform, TEMP_TRANSLATION, TEMP_SCALE);
            this.clearChangedStates();
        }
        return this.transform;
    }
    
    /**
     * Liefert die x-Koordinate der linken oberen Ecke der uebergeordneten
     * Dimension, falls vorhanden; sonst <tt>0</tt>.
     * @return Die x-Koordinate
     */    
    private int getParentGlobalX() {
        return this.node.getParentNode() == null ? 0 : this.node.getParentNode().getDimension().getGlobalX();
    }
    
    /**
     * Liefert die y-Koordinate der linken oberen Ecke der uebergeordneten
     * Dimension, falls vorhanden; sonst <tt>0</tt>.
     * @return Die y-Koordinate
     */
    private int getParentGlobalY() {
        return this.node.getParentNode() == null ? 0 : this.node.getParentNode().getDimension().getGlobalY();
    }
    
    /**
     * Liefert die Breite der uebergeordneten Dimension, falls vorhanden; sonst 
     * <tt>0</tt>.
     * @return Die Breite der uebergeordneten Diemnsion
     */
    private int getParentWidth() {
        return this.node.getParentNode() == null ? 0 : this.node.getParentNode().getDimension().width;
    }
    
    /**
     * Liefert die Hoehe der uebergeordneten Dimension, falls vorhanden; sonst
     * <tt>0</tt>.
     * @return Die Hoehe der uebergeordneten Diemnsion
     */
    private int getParentHeight() {
        return this.node.getParentNode() == null ? 0 : this.node.getParentNode().getDimension().height;
    }
    
    /**
     * Liefert die globale x-Koordinate der linken oberen Ecke dieser Dimension.
     * @return Die globale x-Koordinate
     */
    public int getGlobalX() {
        switch(this.referencePoint & REF_H_MASK) {
            case REF_H_L:
                return this.getParentGlobalX() + this.posX;
            case REF_H_R:
                return this.getParentGlobalX() + this.posX + this.getParentWidth() - this.width;
            case REF_H_C:
                return this.getParentGlobalX() + this.posX + this.getParentWidth() / 2 - this.width / 2;
            default:
                Logger.INSTANCE.warning("Illegal reference point.");
                return 0;
        }
    }
    
    /**
     * Liefert die globale y-Koordinate der linken oberen Ecke dieser Dimension.
     * @return Die globale y-Koordinate
     */
    public int getGlobalY() {
        switch(this.referencePoint & REF_V_MASK) {
            case REF_V_T:
                return this.getParentGlobalY() + this.posY;
            case REF_V_B:
                return this.getParentGlobalY() + this.posY + this.getParentHeight() - this.height;
            case REF_V_C:
                return this.getParentGlobalY() + this.posY + this.getParentHeight() / 2 - this.height / 2;
            default:
                Logger.INSTANCE.warning("Illegal reference point.");
                return 0;
        }
    }
    
    /**
     * Prueft, ob sich ein Punkt innerhalb des von dieser Dimension
     * beschriebenen Rechtecks befindet.
     * @param globalX Die x-Koordinate des Punktes
     * @param globalY Die y-Koordinate des Punktes
     * @return <tt>true</tt>, falls sich der Punkt innerhalb des Rechtecks
     *         befindet
     */
    public boolean contains(int globalX, int globalY) {
        int thisGlobalX = this.getGlobalX();
        int thisGlobalY = this.getGlobalY();
        return thisGlobalX <= globalX && globalX < thisGlobalX + this.width &&
                thisGlobalY <= globalY && globalY < thisGlobalY + this.height;
    }
    
    /**
     * Wandelt eine x-Koordinate eines Pixels in eine x-Koordinate des
     * Bildschirms um. Ein Rechteck [0,resX[&times;[0,resY[ wird dabei auf
     * [-1,1]&sup2 abgebildet.
     * @param x Die x-Koordinate des Pixels
     * @return Die x-Koordinate des Bildschirms
     */
    public float resToScreenX(int x) {
        float screenPercentage = (float)x / (float)GameApp.getGraphicsLayer().getWidth();
        return 2.0f * screenPercentage - 1.0f;
    }
    
    /**
     * Wandelt eine y-Koordinate eines Pixels in eine y-Koordinate des
     * Bildschirms um. Ein Rechteck [0,resX[&times;[0,resY[ wird dabei auf
     * [-1,1]&sup2 abgebildet.
     * @param y Die y-Koordinate des Pixels
     * @return Die y-Koordinate des Bildschirms
     */
    public float resToScreenY(int y) {
        float screenPercentage = (float)y / (float)GameApp.getGraphicsLayer().getHeight();
        return -2.0f * screenPercentage + 1.0f;
    }
    
    /**
     * Schreibt in einen Vektor die beiden Punkte, die das Clipping Rechteck
     * dieser Dimension beschreiben.
     * @param dest Der Vektor
     */
    public void getClippingRect(Vector4f dest) {
        dest.x = this.resToScreenX(this.getGlobalX());
        dest.y = this.resToScreenY(this.getGlobalY());
        dest.z = this.resToScreenX(this.getGlobalX() + this.getWidth());
        dest.w = this.resToScreenX(this.getGlobalY() + this.getHeight());
    }
}
