/*
 * This software is the proprietary information of nico3000.
 */
package n3gui;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.font.LineBreakMeasurer;
import java.awt.font.TextAttribute;
import java.awt.font.TextLayout;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.text.AttributedString;
import java.util.EnumMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import logging.Logger;
import gameapp.GameApp;
import static n3gui.DrawHelper.ALIGN_H_C;
import static n3gui.DrawHelper.ALIGN_V_C;
import resource.ResHandle;
import resource.Resource;
import resource.loader.ImageExtraData;

/**
 *
 * @author nico3000
 */
public class DrawHelper {
    public static enum ColorType {
        /**
         * Standard Hintergrundfarbe
         */
        BACKGROUND,
        /**
         * Standard Rahmenfarbe.
         */
        BORDER,
        /**
         * Standard Schriftfarbe.
         */
        FONT,
    }
    
    /**
     * Horizontal links aurichten.
     */
    public static final int ALIGN_H_L = 1;  // ...0001
    /**
     * Horizontal zentriert aurichten.
     */
    public static final int ALIGN_H_C = 2;  // ...0010
    /**
     * Horizontal rechts aurichten.
     */
    public static final int ALIGN_H_R = 3;  // ...0011
    /**
     * Vertikal oben aurichten.
     */
    public static final int ALIGN_V_T = 4;  // ...0100
    /**
     * Vertikal zentriert aurichten.
     */
    public static final int ALIGN_V_C = 8;  // ...1000
    /**
     * Vertikal unten aurichten.
     */
    public static final int ALIGN_V_B = 12; // ...1100
    /**
     * Bitmaske fuer die horizontale Ausrichtung.
     */
    private static final int ALIGN_H_MASK = 3;
    /**
     * Bitmaske fuer die vertikale Ausrichtung.
     */
    private static final int ALIGN_V_MASK = 12;
    
    private final Map<ColorType, Color> defaultColors = new EnumMap<>(ColorType.class);
    private int defaultBorderRadius = 32;
    private final Font defaultFont = new Font("Verdana", Font.PLAIN, 16);
    
    public DrawHelper(Map<ColorType, Color> defaultColors) {
        this.defaultColors.put(ColorType.BACKGROUND, new Color(20, 25, 30, 223));
        this.defaultColors.put(ColorType.BORDER, new Color(63, 63, 63, 255));
        this.defaultColors.put(ColorType.FONT, new Color(223, 223, 223, 255));
        if(defaultColors != null) {
            for(ColorType key : defaultColors.keySet()) {
                this.defaultColors.put(key, defaultColors.get(key));
            }
        }
    }
    
    public void setDefaultColor(ColorType type, Color color) {
        this.defaultColors.put(type, color);
    }
    
    public Color getDefaultColor(ColorType type) {
        return this.defaultColors.get(type);
    }

    public Font getDefaultFont() {
        return this.defaultFont;
    }
    
    public void drawText(String text, Graphics2D target, Rectangle bounds, int alignment, int shadowOffset) {
        AttributedString attrString = new AttributedString(text);
        attrString.addAttribute(TextAttribute.FONT, target.getFont());
        LineBreakMeasurer measurer = new LineBreakMeasurer(attrString.getIterator(), target.getFontRenderContext());
        List<TextLayout> lines = new LinkedList<>();
        while(measurer.getPosition() < text.length()) {
            int nextLineBreak = text.indexOf('\n', measurer.getPosition());
            lines.add(nextLineBreak == -1 ? measurer.nextLayout(bounds.width) : measurer.nextLayout(bounds.width, nextLineBreak, false));
        }
        Shape clip = target.getClip();
        Color color = target.getColor();
        //target.setClip(bounds.x, bounds.y, bounds.width, bounds.height);
        int currentY = bounds.y + target.getFontMetrics().getAscent();
        int extentY = lines.size() * target.getFontMetrics().getHeight();
        switch(alignment & ALIGN_V_MASK) {
            case ALIGN_V_C:
                currentY += (bounds.height - extentY) / 2;
                break;
            case ALIGN_V_B:
                currentY += bounds.height - extentY;
                break;
        }
        for(TextLayout layout : lines) {
            int x = bounds.x;
            switch(alignment & ALIGN_H_MASK) {
                case ALIGN_H_R:
                    x += bounds.width - (int)layout.getBounds().getWidth();
                    break;
                case ALIGN_H_C:
                    x += (bounds.width - (int)layout.getBounds().getWidth()) / 2;
                    break;                    
            }
            target.setColor(Color.BLACK);
            layout.draw(target, x + shadowOffset, currentY + shadowOffset);
            target.setColor(color);
            layout.draw(target, x, currentY);
            currentY += target.getFontMetrics().getHeight();
        }
        target.setClip(clip);
    }
    
    public BufferedImage getImage(String resource) {
        ResHandle handle = GameApp.getResCache().getHandle(new Resource(resource));
        if(handle == null || !(handle.getExtra() instanceof ImageExtraData)) {
            Logger.INSTANCE.warning("Invalid image resource %s.", resource);
            return null;
        } else {
            ImageExtraData data = (ImageExtraData)handle.getExtra();
            BufferedImage image = new BufferedImage(data.getWidth(), data.getHeight(), BufferedImage.TYPE_4BYTE_ABGR);
            InputStream stream = handle.getBufferAsInputStream();
            byte floatBytes[] = new byte[4];
            ByteBuffer buffer = ByteBuffer.wrap(floatBytes);
            for(int y=0; y < data.getHeight(); ++y) {
                for(int x=0; x < data.getWidth(); ++x) {
                    float color[] = { 1.0f, 1.0f, 1.0f, 1.0f };
                    for(int c=0; c < data.getChannels(); ++c) {
                        try {
                            stream.read(floatBytes);
                        } catch (IOException ex) {
                            Logger.INSTANCE.error(ex.getMessage());
                            return null;
                        }
                        color[c] = buffer.getFloat(0);
                    }
                    image.setRGB(x, y, new Color(color[0], color[1], color[2], color[3]).getRGB());
                }
            }
            return image;
        }
    }
    
    public BufferedImage createBaseImage(int width, int height) {
        return new BufferedImage(width, height, BufferedImage.TYPE_4BYTE_ABGR);
    }
    
    public void drawDecoration(Graphics2D target, Rectangle bounds) {
        target.setBackground(new Color(0, 0, 0, 0));
        target.setColor(this.getDefaultColor(ColorType.BACKGROUND));
        target.fillRoundRect(bounds.x, bounds.y, bounds.width, bounds.height, this.defaultBorderRadius, this.defaultBorderRadius);
        target.setColor(this.getDefaultColor(ColorType.BORDER));
        target.drawRoundRect(bounds.x, bounds.y, bounds.width, bounds.height, this.defaultBorderRadius, this.defaultBorderRadius);
    }
}
