/*
 * This software is the proprietary information of nico3000.
 */
package n3gui;

import graphics.shader.UniformBuffer;
import java.awt.Rectangle;
import java.nio.FloatBuffer;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import logging.Logger;
import org.lwjgl.BufferUtils;
import org.lwjgl.util.vector.Vector4f;

/**
 *
 * @author nico3000
 */
public class GuiTreeNode implements Comparable<GuiTreeNode> {
    /**
     * Die ID des Knotens. Ist in einem <tt>GuiTree</tt> eindeutig.
     */
    private final String id;
    /**
     * Der Name des Knotens.
     */
    private String name;
    /**
     * Der Vaterknoten, falls vorhanden.
     */
    private GuiTreeNode parent = null;
    /**
     * Die raeumliche Ausdehnung des Knotens.
     */
    private final Dimension dimension = new Dimension(this);
    /**
     * Die Gui Komponente des Knotens.
     */
    private final Component component;
    /**
     * Liste alle Kindknoten, sortiert nach ihrer Tiefe.
     */
    private final List<GuiTreeNode> children = new LinkedList<>();
    /**
     * Gibt an, ob das Element sichtbar ist.
     */
    private boolean visible = true;
    /**
     * Die Tiefe des Elements. Tiefe Elemente liegen hinter hoeheren.
     */
    private int zOrder = 0;
    /**
     * Gibt an, ob dieses Element immer ueber allen anderen gezeichnet werden
     * soll.
     */
    private boolean alwaysOnTop = false;
    /**
     * Gibt an, ob das Element gehovert ist.
     */
    private boolean hovered = false;
    /**
     * Gibt an, ob das Element fokussiert ist.
     */
    private boolean focused = false;
    /**
     * Gibt an, ob sich der Fokusstatus veraendert hat.
     */
    private boolean focusChanged = false;
    /**
     * Gibt an, ob sich der Hoverstatus veraendert hat.
     */
    private boolean hoverChanged = false;
    /**
     * Farbe des Knotens fuer die schematische Debugausgabe.
     */
    private final Vector4f debugColor = new Vector4f((float)Math.random(), (float)Math.random(), (float)Math.random(), 1.0f);
    /**
     * Die Daten des Transform Uniform Buffers.
     */
    private final FloatBuffer transformData = BufferUtils.createFloatBuffer(24);
    /**
     * Der Transform Uniform Buffer.
     */
    private final UniformBuffer transformBuffer = new UniformBuffer();
    /**
     * Der zugehoerige Baum.
     */
    private final GuiTree tree;

    /**
     * Konstruktor mit ID und Gui Komponente.
     * @param id Die ID
     * @param component Die Gui Komponente
     */
    GuiTreeNode(String id, Component component, GuiTree tree) {
        this.id = id == null ? ("node_" + this.hashCode()) : id;
        this.component = component;
        this.tree = tree;
        this.name = this.id;
    }

    /**
     * Liefert die ID.
     * @return Die ID
     */
    public String getId() {
        return id;
    }

    /**
     * Liefert den Namen.
     * @return Der Name
     */
    public String getName() {
        return name;
    }
    
    /**
     * Liefert die Dimensionkomponente des Knotens.
     * @return Die Dimension
     */
    public Dimension getDimension() {
        return this.dimension;
    }
    
    /**
     * Liefert die Komponente des Knotens.
     * @return Die Komponente
     */
    public Component getComponent() {
        return this.component;
    }

    /**
     * Setzt einen neuen Namen.
     * @param name Der neue Name
     */
    public void setName(String name) {
        this.name = name;
    }
    
    /**
     * Liefert den zugehoerigen Baum.
     * @return Der Baum
     */
    public GuiTree getTree() {
        return this.tree;
    }

    /**
     * Liefert, ob das Element sichtbar ist.
     * @return <tt>true</tt>, falls sichtbar
     */
    public boolean isVisible() {
        return this.visible;
    }

    /**
     * Setzt die Sichtbarkeit des Elements.
     * @param visible Die Sichtbarkeit
     */
    public void setVisible(boolean visible) {
        this.visible = visible;
    }

    /**
     * Liefert die Tiefe des Elements.
     * @return Die Tiefe
     */
    public int getzOrder() {
        return zOrder;
    }

    /**
     * Setzt die Tiefe des Elements. Tiefe Element liegen hinter hoeheren.
     * @param zOrder Die neue Tiefe
     */
    public void setzOrder(int zOrder) {
        this.zOrder = zOrder;
    }

    /**
     * Liefert, ob dieses Element ueber allen anderen gezeichnet werden soll.
     * @return <tt>true</tt>, falls das Element ueber allen anderen gezeichnet
     *         werden soll
     */
    public boolean isAlwaysOnTop() {
        return alwaysOnTop;
    }

    /**
     * Setzt, ob dieses Element ueber allen anderen gezeichnet werden soll.
     * @param alwaysOnTop <tt>true</tt>, falls das Element ueber allen anderen
     *                    gezeichnet werden soll
     */
    public void setAlwaysOnTop(boolean alwaysOnTop) {
        this.alwaysOnTop = alwaysOnTop;
        this.dimension.positionChanged();
    }

    /**
     * Liefert, ob der Cursor sich ueber dem Element befindet.
     * @return <tt>true</tt>, falls sich der Cursor darueber befindet
     */
    public boolean isHovered() {
        return this.hovered;
    }

    /**
     * Setzt, ob der Cursor auf dem Element ist und nicht.
     * @param focused <tt>true</tt>, falls der Cursor ueber dem Element ist
     * @return <tt>true</tt>, falls das Element auf das Event reagiert
     */
    public boolean setHovered(boolean hovered) {
        boolean accepted = this.component != null && this.component.isHoverable();
        if(accepted && this.hovered != hovered) {
            this.hoverChanged = !this.hoverChanged;
            this.hovered = hovered;
        }
        return accepted;
    }

    /**
     * Liefert, ob das Element den Fokus hat.
     * @return <tt>true</tt>, falls es den Fokus hat
     */
    public boolean isFocused() {
        return this.focused;
    }

    /**
     * Setzt, ob das Element den Fokus hat.
     * @param focused <tt>true</tt>, falls das Element den Fokus erhalten hat
     * @return <tt>true</tt>, falls das Element auf das Event reagiert
     */
    public boolean setFocused(boolean focused) {
        boolean accepted = this.component != null && this.component.isFocusable();
        if(accepted && this.focused != focused) {
            this.focusChanged = !this.focusChanged;
            this.focused = focused;
        }
        if(this.parent != null) {
            this.parent.setFocused(focused);
        }   
        return accepted;
    }
    
    /**
     * Liefert den Qualifier dieses Knotens.
     * @return Der Qualifier
     */
    public String getQualifier() {
        return (this.parent == null || this.parent.id.equals("root")) ? this.name : String.format("%s.%s", this.parent.getQualifier(), this.name);
    }
    
    /**
     * Initialisiert den Node.
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean init() {
        if(!this.transformBuffer.init()) {
            Logger.INSTANCE.warning("Error while initializing uniform buffer of %s.", this);
            return false;
        }
        this.transformBuffer.setBindingIndex(UniformBuffer.BINDING_MODEL);
        if(this.component != null) {
            this.dimension.setSize(this.component.getPreferredWidth(), this.component.getPreferredHeight());
        }
        return true;
    }
    
    /**
     * Zerstoert diesen Knoten und alle Nachfolger.
     */
    void destroy() {
        if(this.component != null) {
            this.component.destroy();
        }
        this.transformBuffer.destroy();
        for(GuiTreeNode child : this.children) {
            child.destroy();
        }
    }
    
    public boolean contains(int x, int y) {
        return this.dimension.contains(x, y) && (this.parent == null || this.parent.contains(x, y));
    }
    
    /**
     * Fuegt diesem Knoten einen Knoten als Kind an. Der Kindknoten darf nicht
     * <tt>null</tt>, nicht der Knoten selbst und nicht sein Vorfahre sein.
     * @param child Der Kindknoten
     * @return <tt>true</tt>, falls erfolgreich
     */
    boolean addChild(GuiTreeNode child) {
        if(child == null) {
            Logger.INSTANCE.warning("Tried to add null as a child to %s.", this);
            return false;
        } else if(child == this) {
            Logger.INSTANCE.warning("Tried to add %s as a child to itself.", child);
            return false;
        } else if(this.isPredecessor(child)) {
            Logger.INSTANCE.warning("Tried to add %s as a child to its predecessor %s.", child, this);
            return false;
        } else {
            if(child.parent != null) {
                if(!child.parent.removeChild(child)) {
                    Logger.INSTANCE.warning("Failed to remove %s from its parent %s.", child, child.parent);
                    return false;
                }
            }
            child.parent = this;
            this.children.add(child);
            return true;
        }
    }
    
    /**
     * Prueft, ob ein gegebener Knoten ein Vorgaenger dieses Knotens ist.
     * @param node Der zu pruefende Vorgaenger
     * @return <tt>true</tt>, falls <tt>node</tt> ein Vorgaenger von dieses
     *         Knotens ist
     */
    private boolean isPredecessor(GuiTreeNode node) {
        if(this.parent == null) {
            return false;
        } else if(this.parent == node) {
            return true;
        } else {
            return this.parent.isPredecessor(node);
        }
    }
        
    /**
     * Entfernt einen Kindknoten.
     * @param child Der Kindknoten
     * @return <tt>true</tt>, falls erfolgreich
     */
    boolean removeChild(GuiTreeNode child) {
        if(child == null) {
            Logger.INSTANCE.warning("Trying to remove null from %s.", this);
            return false;
        } else if(!this.children.contains(child)) {
            Logger.INSTANCE.warning("Tried to remove non-child %s from %s.", child, this);
            return false;
        } else {
            child.parent = null;
            child.dimension.parentChanged();
            this.children.remove(child);
            return true;
        }
    }
    
    /**
     * Sucht einen direkten Kindknoten anhand seines Namens.
     * @param name Der Name des Kindknotens
     * @return Der Kindknoten oder <tt>null</tt>, falls es keinen mit dem
     *         gegebenen Namen gibt
     */
    public GuiTreeNode findChildByName(String name) {
        for(GuiTreeNode child : this.children) {
            if(child.name.equals(name)) {
                return child;
            }
        }
        return null;
    }
    
    /**
     * Liefert eine nicht veraenderbare Liste der Kindknoten.
     * @return Die Liste
     */
    List<GuiTreeNode> getChildren() {
        return Collections.unmodifiableList(this.children);
    }
    
    /**
     * Liefert den Vaterknoten, falls vorhanden; sonst <tt>null</tt>.
     * @return Der Vaterknoten
     */
    public GuiTreeNode getParentNode() {
        return this.parent;
    }
    
    /**
     * Falls sich die Dimension geaendert hat, wird der Uniform Buffer
     * aktualisiert.
     */
    private void updateUniform() {
        if(this.dimension.hasPosChanged() || this.dimension.hasSizeChanged()) {
            this.transformData.position(0);
            this.dimension.getTransform().store(this.transformData);
            this.debugColor.store(this.transformData);
            Vector4f clippingRect = new Vector4f(-1.0f, -1.0f, 1.0f, 1.0f);
            if(this.parent != null && this.component != null && !this.isAlwaysOnTop()) {
                int levelsToGo = this.component.getClipOffset();
                GuiTreeNode pre = this.parent;
                while(levelsToGo > 0 && pre != null) {
                    pre = pre.parent;
                    --levelsToGo;
                }
                if(pre != null && pre.component != null) {
                    Rectangle parentClippingRect = pre.component.getClippingRect();
                    clippingRect.x = this.parent.dimension.resToScreenX(parentClippingRect.x);
                    clippingRect.y = this.parent.dimension.resToScreenY(parentClippingRect.y + parentClippingRect.height);
                    clippingRect.z = this.parent.dimension.resToScreenX(parentClippingRect.x + parentClippingRect.width);
                    clippingRect.w = this.parent.dimension.resToScreenY(parentClippingRect.y);
                }
            }
            clippingRect.store(this.transformData);
            this.transformData.position(0);
            this.transformBuffer.setData(this.transformData);
        }
    }
    
    /**
     * Zeichnet die Umrisse der Elements mit einer Frbe gefuellt.
     */
    public void drawSchemes() {
        if(this.isVisible()) {
            if(this.component != null && this.component.bindTexture(GuiTree.EFFECT_TEX_SLOT)) {
                this.updateUniform();
                this.transformBuffer.bind();
                GuiTree.QUAD.draw();
            }
            for(GuiTreeNode child : this.children) {
                child.drawSchemes();
            }
        }
    }
    
    /**
     * Updatet diesen und alle Kindknoten.
     * @param deltaMillis Die Dauer des letzten Frames in Millisekunden
     */
    void update(long deltaMillis) {
        if(this.component != null) {
            if(this.focusChanged) {
                this.component.repaintIfBitIsSet(Component.REPAINT_ON_FOCUS_CHANGED);
                this.component.onFocusChanged();
                this.focusChanged = false;
            }
            if(this.hoverChanged) {
                this.component.repaintIfBitIsSet(Component.REPAINT_ON_HOVER_CHANGED);
                this.component.onHoverChanged();
                this.hoverChanged = false;
            }
            if(this.dimension.hasPosChanged()) {
                this.component.onMoved();
            }
            if(this.dimension.hasSizeChanged()) {
                this.component.onSizeChanged();
            }
            this.component.onUpdate(deltaMillis);
        } else {
            this.dimension.clearChangedStates();
        }
        boolean sorted = true;
        GuiTreeNode lastChild = null;
        for(GuiTreeNode child : this.children) {
            child.update(deltaMillis);
            if(lastChild != null && lastChild.compareTo(child) > 0) {
                sorted = false;
            }
            lastChild = child;
        }
        if(!sorted) {
            Logger.INSTANCE.info("Sorting children of %s.", this);
            Collections.sort(this.children);
        }
    }
    
    @Override
    public String toString() {
        if(this.component == null) {
            return String.format("<no component> '%s' (ID: %s)", this.name, this.id);
        } else {
            return String.format("%s '%s' (ID: %s)", this.component.getClass().getSimpleName(), this.name, this.id);
        }
    }

    @Override
    public int compareTo(GuiTreeNode o) {
        // null nicht erlaubt
        if(o == null) {
            return 1;
        }
        // zuerst die alwaysOnTop elemente
        if(this.alwaysOnTop && !o.alwaysOnTop) {
            return 1;
        } else if(!this.alwaysOnTop && o.alwaysOnTop) {
            return -1;
        }
        // dann die gehoverten
        if(this.focused && !o.focused) {
            return 1;
        } else if(!this.focused && o.focused) {
            return -1;
        }
        // dann nach zOrder
        return Integer.compare(this.zOrder, o.zOrder);
    }
}
