/*
 * Copyright (C) 2013 nico3000 - All Rights Reserved
 */
package resource;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import logging.Logger;

/**
 * Ein Handle auf eine Resource enthaelt den Buffer auf die rohen oder
 * bearbeiteten Daten.
 * @author nico3000
 */
public final class ResHandle {
    /**
     * Der Status eines Handles.
     */
    public static enum Status {
        /**
         * Wartet darauf, geladen zu werden.
         */
        PENDING,
        /**
         * Daten sind geladen und bereit.
         */
        LOADED,
        /**
         * Laden der Daten fehlgeschlagen.
         */
        WONT_LOAD,
    }
    
    /**
     * Die Resource des Handles.
     */
    private final Resource resource;
    /**
     * Die Daten der Resource.
     */
    private byte buffer[];
    /**
     * Die Groesse der Resource.
     */
    private int size;
    /**
     * Extradaten der Resource.
     */
    private Object extra;
    /**
     * Der aktuelle Status des Handles.
     */
    private Status status;
    /**
     * Ist das Handle gelockt, koennen die Daten nicht geloescht werden.
     */
    private boolean locked = false;
    /**
     * Zeitstempel, als das Handle erstellt wurde.
     */
    private long created = System.currentTimeMillis();

    /**
     * Konstruktor.
     * @param resource Die Resource
     * @param buffer Der Buffer, der die Daten enthaelt
     * @param resCache Der zugehoerige Resource Cache
     */
    ResHandle(Resource resource) {
        this.resource = resource;
        this.extra = null;
        this.buffer = null;
        this.size = 0;
        this.status = Status.PENDING;
    }
    
    /**
     * Setzt den Buffer.
     * @param buffer Der Buffer
     */
    void setBuffer(byte buffer[]) {
        if(this.buffer != null) {
            Logger.INSTANCE.error("cannot set buffer of ResHandle twice");
            return;
        }
        this.buffer = buffer;
        this.size = this.buffer.length;
        float loadTime = 1e-3f * (float)(System.currentTimeMillis() - this.created);
        //Logger.INSTANCE.info(String.format("resource loaded in %.2fs: %s (%s)", loadTime, this.resource.getName(), MathUtil.byteToString(this.size)));
    }

    /**
     * Liefert die Resource.
     * @return Die Resource
     */
    Resource getResource() {
        return this.resource;
    }
    
    /**
     * Loescht die unterliegenden Daten.
     */
    void destroy() {
        this.buffer = null;
        //Logger.INSTANCE.info("resource deleted: " + this.resource.getName());
    }

    /**
     * Liefert die Rohdaten der Resource. Diese Methode erhoeht den Reference
     * Counter automatisch um 1. Wird der Buffer nicht weiter benutzt, so muss
     * explizit <tt>delRef()</tt> aufgerufen werden.
     * @return Die Rohfaten
     */
    public byte[] getBuffer() {
        return this.buffer;
    }
    
    /**
     * Liefert einen InputStream der Rohdaten der Resource. Diese Methode
     * erhoeht den Reference Counter automatisch um 1. Wird der Buffer nicht
     * weiter benutzt, so muss explizit <tt>delRef()</tt> aufgerufen werden.
     * @return Ein InputStream mit den Rohdaten
     */
    public InputStream getBufferAsInputStream() {
        return new ByteArrayInputStream(this.getBuffer());
    }

    /**
     * Liefert die Groesse der unterliegenden Daten in Byte.
     * @return Die Groesse in Byte
     */
    public int getSize() {
        return this.size;
    }

    /**
     * Liefert einen veraenderbaren Buffer, der die Daten enthaelt. Diese
     * Methode erhoeht den Reference Counter automatisch um 1. Wird der Buffer
     * nicht weiter benutzt, so muss explizit <tt>delRef()</tt> aufgerufen
     * werden.
     * @return Ein veraenderbarer Buffer
     */
    public byte[] getWritableBuffer() {
        return this.buffer;
    }

    /**
     * Liefert die Extradaten dieser Resource.
     * @return Die Extradaten
     */
    public Object getExtra() {
        return extra;
    }

    /**
     * Setzt die Extradaten der Resource.
     * @param extra Die neuen Extradaten
     */
    public void setExtra(Object extra) {
        this.extra = extra;
    }

    /**
     * Liefert den Status des Handles.
     * @return Der Status
     */
    synchronized Status getStatus() {
        return this.status;
    }
    
    /**
     * Schuetzt das Handle vor dem Loeschen.
     */
    synchronized void lock() {
        this.locked = true;
    }

    synchronized boolean isLocked() {
        return locked;
    }
    
    /**
     * Gibt das Handle zum Loeschen frei.
     */
    public void unlock() {
        this.locked = false;
    }
    
    /**
     * Setzt den Status des Handles.
     * @param status Der Status
     */
    synchronized void setStatus(Status status) {
        if(status == Status.PENDING) {
            Logger.INSTANCE.error("setting status of ResHandle to PENDING is not allowed");
        } else {
            this.status = status;
            this.notifyAll();
        }
    }
    
    /**
     * Wartet solange, bis die Resource geladen ist.
     * @return <tt>true</tt>, falls die Resource erfolgreichgeladen wurde.
     */
    synchronized boolean waitForFinalStatus() {
        if(this.getStatus() == Status.PENDING) {
            try {
                this.wait();
            } catch (InterruptedException ex) {
                Logger.INSTANCE.error(ex.getMessage());
            }
        }
        return this.status == Status.LOADED;
    }
}
