/*
 * Copyright (C) 2013 nico3000 - All Rights Reserved
 */
package resource;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Queue;
import logging.Logger;
import resource.file.ResourceFile;
import resource.loader.DefaultResourceLoader;
import resource.loader.DstImageLoader;
import resource.loader.ImageLoader;
import resource.loader.ResourceLoader;
import util.MathUtil;

/**
 * Implementiert einen Resource Cache, mit dessen Hilfe Daten von der Festplatte
 * verwaltet werden koennen.
 * @author nico3000
 */
public class ThreadedResCache {
    /**
     * Die maximale Groesse des Caches in Byte.
     */
    private final long maxSize;
    /**
     * Die aktuelle Groesse des Caches in Byte.
     */
    private long currentSize;
    /**
     * Die Quelle der Daten.
     */
    private ResourceFile file;
    /**
     * Die eingetragenen ResourceLoader, die Rohdaten weiterverarbeiten koennen.
     */
    private final List<ResourceLoader> loaders = new LinkedList<>();
    /**
     * Alle momentan bekannten Resourcen.
     */
    private final Map<String, ResHandle> resources = new HashMap<>();
    /**
     * Je weiter hinten ein Handle in der Liste steht, desto kuerzlicher wurde
     * es verwendet.
     */
    private final LinkedList<ResHandle> lru = new LinkedList<>();
    /**
     * Die Schlange der noch zu ladenden Resourcen.
     */
    private final Queue<ResHandle> pending = new LinkedList<>();
    /**
     * Der Thread, der die Daten laedt.
     */
    private final ResourceLoaderThread thread = new ResourceLoaderThread();

    /**
     * Erzeugt einen neuen Resource Cache mit fester maximaler Groesse.
     * @param sizeInMb Die Groesse in MiB
     */
    public ThreadedResCache(int sizeInMb) {
        this.maxSize = 1024L * 1024L * (long)sizeInMb;
        this.currentSize = 0;
    }

    /**
     * Setzt die Quelle der Daten.
     * @param file Die Datei
     */
    public void setFile(ResourceFile file) {
        if(this.file != null) {
            Logger.INSTANCE.error("cannot set file for ThreadedResCache twice");
            return;
        }
        this.file = file;
    }
    
    /**
     * Initialisiert den Resource Cache.
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean init() {
        if(this.file != null && this.file.open()) {
            this.loaders.add(new ImageLoader());
            this.loaders.add(new DstImageLoader());
            this.loaders.add(new DefaultResourceLoader());
            this.thread.setDaemon(true);
            this.thread.start();
            return true;
        } else {
            return false;
        }
    }
    
    /**
     * Zerstoert den Resource Cache und alle mit ihm geladenen Daten,
     * unabhaenigig davon, ob die Handle gelockt sind und nicht.
     */
    public void destroy() {
        this.thread.active = false;
        synchronized(this.pending) {
            this.pending.clear();
            this.pending.notifyAll();
        }
        try {
            this.thread.join();
        } catch (InterruptedException ex) {
            Logger.INSTANCE.error(ex.getMessage());
        }
        for(ResHandle handle : this.lru) {
            if(handle.isLocked()) {
                Logger.INSTANCE.warning("Resource %s is still locked.", handle.getResource().getName());
            }
        }
        this.file.close();
    }
    
    /**
     * Fuegt eine Resourcenanfrage hinzu, damit diese geladen werden kann, bevor
     * sie benoetigt wird. Ein späterer Aufruf von <tt>getHandle()</tt> liefert
     * das Ergebnis potentiell schneller.
     * @param resource Die Resource
     */
    public void addRequest(Resource resource) {
        if(!this.resources.containsKey(resource.getName())) {
            ResHandle handle = new ResHandle(resource);
            this.resources.put(resource.getName(), handle);
            synchronized(this.pending) {
                this.pending.add(handle);
                this.pending.notifyAll();
            }
        }
    }
    
    /**
     * Liefert einen Handle auf eine Resource und lockt diese. <b>HINWEIS</b>:
     * Nachdem der Handle nicht mehr benoetigt wird, unbedingt <tt>unlock()</tt>
     * darauf aufrufen!
     * @param resource Die zu ladende Resource
     * @return Ein Handle auf die Resource oder <tt>null</tt>, falls beim Laden
     *         ein Fehler aufgetreten ist
     */
    public ResHandle getHandle(Resource resource) {
        ResHandle handle;
        synchronized(this) {
            handle = this.resources.get(resource.getName());
            if(handle == null) {
                this.addRequest(resource);
                handle = this.resources.get(resource.getName());
            }
            handle.lock();
        }
        if(!handle.waitForFinalStatus()) {
            this.resources.remove(resource.getName());
            return null;
        }
        synchronized(this) {
            this.lru.remove(handle);
            this.lru.addLast(handle);
        }
        return handle;
    }
    
    /**
     * Loescht eine Resource explizit, sofern sie geladen ist.
     * @param resource Die Resource
     */
    public synchronized void deleteResource(Resource resource) {
        ResHandle handle = this.resources.get(resource.getName());
        if(handle != null && handle.getStatus() == ResHandle.Status.LOADED) {
            this.currentSize -= handle.getSize();
            this.lru.remove(handle);
            this.resources.remove(resource.getName());
            handle.destroy();
        }
    }
    
    /**
     * Wartet auf einen neuen Ladeauftrag.
     * @return Ein neuer Auftrag oder <tt>null</tt>
     */
    private ResHandle getPendingHandle() {
        synchronized(this.pending) {
            if(this.pending.isEmpty()) {
                try {
                    this.pending.wait();
                } catch (InterruptedException ex) {
                    Logger.INSTANCE.error(ex.getMessage());
                }
            }
            return this.pending.poll();
        }
    }
    
    /**
     * Laedt die zu einem Handle zugehoerigen Daten.
     * @param handle Der Handle
     * @return <tt>true</tt>, falls erfolgreich
     */
    private boolean load(ResHandle handle) {
        if(handle == null) {
            return false;
        }
        ResourceLoader loader = null;
        for(ResourceLoader loaderIt : this.loaders) {
            if(handle.getResource().getName().matches(loaderIt.getPattern())) {
                loader = loaderIt;
                break;
            }
        }
        if(loader == null) {
            Logger.INSTANCE.error("no default loader");
            return false;
        }
        int rawSize = this.file.getRawResourceSize(handle.getResource());
        if(loader.useRawFile()) {
            if(!this.freeEnoughMemory(rawSize)) {
                Logger.INSTANCE.error("not enough space for resource " + handle.getResource().getName());
                return false;
            }
            byte rawBuffer[] = new byte[rawSize];
            this.file.getRawResource(handle.getResource(), rawBuffer);
            handle.setBuffer(rawBuffer);
        } else {
            byte rawBuffer[] = new byte[rawSize];
            this.file.getRawResource(handle.getResource(), rawBuffer);
            int size = loader.getLoadedResourceSize(rawBuffer, rawSize);
            if(!this.freeEnoughMemory(size)) {
                Logger.INSTANCE.error("not enough space for resource " + handle.getResource().getName());
                return false;
            }
            handle.setBuffer(new byte[size]);
            if(!loader.loadResource(rawBuffer, rawSize, handle)) {
                Logger.INSTANCE.error("resource " + handle.getResource().getName() + " could not be loaded");
                return false;
            }
        }
        this.currentSize += handle.getSize();
        return true;
    }
    
    /**
     * Loescht solange nicht gelockte Resourcen, bis mindestens <tt>bytes</tt>
     * Byte Speicherplatz frei ist.
     * @param bytes Die Mindestgroesse der freien Bytes
     * @return <tt>true</tt>, falls erfolgreich
     */
    private synchronized boolean freeEnoughMemory(int bytes) {
        if(this.currentSize + bytes < this.maxSize) {
            return true;
        }
        ListIterator<ResHandle> iter = this.lru.listIterator();
        while(iter.hasNext()) {
            ResHandle handle = iter.next();
            if(!handle.isLocked()) {
                Logger.INSTANCE.info("Resource %s deleted.", handle.getResource().getName());
                this.currentSize -= handle.getSize();
                iter.remove();
                if(this.currentSize + bytes < this.maxSize) {
                    return true;
                }
            }
        }
        return false;
    }

    synchronized void memoryHasBeenFreed(int size) {
        this.currentSize -= size;
    }
    
    public void printStats() {
        float usage = 100.0f * ((float)this.currentSize / (float)this.maxSize);
        System.out.println("ThreadedResCache stats:\n----------------------");
        System.out.printf("Usage: %s / %s (%.2f%%)\n", MathUtil.byteToString(this.currentSize), MathUtil.byteToString(this.maxSize), usage);
        for(ResHandle handle : this.resources.values()) {
            String size = MathUtil.byteToString(handle.getSize());
            System.out.printf("- size: %s;\tlocked: %b;\t%s\n", size, handle.isLocked(), handle.getResource().getName());
        }
    }
    
    /**
     * Der Thread laeuft permament im Hintergrund und laedt Resourcen, die
     * angefragt werden.
     */
    private class ResourceLoaderThread extends Thread {
        /**
         * Zeigt an, ob der Thread aktiv sein soll.
         */
        private boolean active = true;

        /**
         * Konstruktor.
         */
        public ResourceLoaderThread() {
            super("ResourceLoaderThread");
        }

        @Override
        public void run() {
            while(this.active) {
                ResHandle handle = ThreadedResCache.this.getPendingHandle();
                if(handle != null) {
                    if(ThreadedResCache.this.load(handle)) {
                        handle.setStatus(ResHandle.Status.LOADED);
                    } else {
                        handle.setStatus(ResHandle.Status.WONT_LOAD);
                    }
                }
            }
        }
        
    }
}
