/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package util;

import com.bulletphysics.linearmath.Transform;
import javax.vecmath.Quat4f;

import logging.Logger;

import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Quaternion;
import org.lwjgl.util.vector.Vector3f;
import org.lwjgl.util.vector.Vector4f;

/**
 * Utilityklasse fuer mathematische Operationen, hauptsaechlich fuer den
 * Datentyp <tt>float</tt>.
 * @author nico3000
 */
public class MathUtil {
    /**
     * Pi in Floatgenauigkeit.
     */
    public static final float PI = (float)Math.PI;
    /**
     * Pi / 2 in Floatgenauigkeit.
     */
    public static final float PI_DIV2 = 0.5f * PI;
    /**
     * Pi / 4 in Floatgenauigkeit.
     */
    public static final float PI_DIV4 = 0.25f * PI;
    /**
     * Pi / 2 in Floatgenauigkeit.
     */
    public static final float PI_MUL2 = 2.0f * PI;
    
    private static final Matrix4f TEMP_MAT = new Matrix4f();
    private static final Matrix4f TEMP_MAT2 = new Matrix4f();
    private static final Vector3f TEMP_VEC3_A = new Vector3f();
    private static final Vector3f TEMP_VEC3_B = new Vector3f();
    private static final Vector3f TEMP_VEC3_C = new Vector3f();
    private static final Vector4f TEMP_VEC4_A = new Vector4f();
    
    /**
     * Schneidet einen Wert so, dass er in einem gegebenen Intervall liegt. Es
     * findet <i>keine</i> Fehlerbehandlung statt!
     * @param val Der Wert
     * @param min Die linke Grenze des Intervalls
     * @param max Die rechte Grenze des Intervalls
     * @return <tt>val</tt>, aber mindestens <tt>min</tt> und hoechstens
     *         <tt>max</tt>
     */
    public static float clamp(float val, float min, float max) {
        return val < min ? min : (val > max ? max : val);
    }
    
    /**
     * Schneidet einen Wert so, dass er in einem gegebenen Intervall liegt. Es
     * findet <i>keine</i> Fehlerbehandlung statt!
     * @param val Der Wert
     * @param min Die linke Grenze des Intervalls
     * @param max Die rechte Grenze des Intervalls
     * @return <tt>val</tt>, aber mindestens <tt>min</tt> und hoechstens
     *         <tt>max</tt>
     */
    public static int clamp(int val, int min, int max) {
        return val < min ? min : (val > max ? max : val);
    }
    
    /**
     * Berechnet den Sinus eines gegebenen Winkels.
     * @param a Der Winkel im Bogenmass
     * @return sin(a)
     */
    public static float sin(float a) {
        return (float)Math.sin(a);
    }
    
    /**
     * Berechnet den Kosinus eines gegebenen Winkels.
     * @param a Der Winkel im Bogenmass
     * @return cos(a)
     */
    public static float cos(float a) {
        return (float)Math.cos(a);
    }
    
    /**
     * Berechnet den Tangens eines gegebenen Winkels.
     * @param a Der Winkel im Bogenmass
     * @return tan(a)
     */
    public static float tan(float a) {
        return (float)Math.tan(a);
    }
    
    public static int posMod(int a, int mod) {
        return ((a % mod) + mod) % mod;
    }
    
    public static float posMod(float a, float mod) {
        return ((a % mod) + mod) % mod;
    }
    
    public static boolean isInRange(float val, float min, float max) {
        return min <= val && val <= max;
    }
    
    /**
     * Erzeugt eine Matrix zur Rotation zm die x-Achse.
     * @param dest Zielmatrix, kann <tt>null</tt> sein
     * @param angle Der Winkel im Bogenmass
     * @return <tt>dest</tt>
     */
    public static Matrix4f rotationX(Matrix4f dest, float angle) {
        if(dest == null) {
            dest = new Matrix4f();
        }
        dest.setIdentity();
        dest.m11 = dest.m22 = cos(angle);
        dest.m21 = -(dest.m12 = sin(angle));
        return dest;
    }
    
    /**
     * Erzeugt eine Matrix zur Rotation zm die y-Achse.
     * @param dest Zielmatrix, kann <tt>null</tt> sein
     * @param angle Der Winkel im Bogenmass
     * @return <tt>dest</tt>
     */
    public static Matrix4f rotationY(Matrix4f dest, float angle) {
        if(dest == null) {
            dest = new Matrix4f();
        }
        dest.setIdentity();
        dest.m00 = dest.m22 = cos(angle);
        dest.m02 = -(dest.m20 = sin(angle));
        return dest;
    }
    
    /**
     * Erzeugt eine Matrix zur Rotation zm die z-Achse.
     * @param dest Zielmatrix, kann <tt>null</tt> sein
     * @param angle Der Winkel im Bogenmass
     * @return <tt>dest</tt>
     */
    public static Matrix4f rotationZ(Matrix4f dest, float angle) {
        if(dest == null) {
            dest = new Matrix4f();
        }
        dest.setIdentity();
        dest.m00 = dest.m11 = cos(angle);
        dest.m10 = -(dest.m01 = sin(angle));
        return dest;
    }
    
    public static float getRotationXFromQuat(Quaternion quat) {
    	return (float)Math.atan2(2 * (quat.x * quat.w + quat.z * quat.y), 1- 2 * (quat.y  * quat.y + quat.x * quat.x));
    }
    
    public static float getRotationYFromQuat(Quaternion quat) {
    	return (float)Math.asin(2 * (quat.w * quat.y - quat.z * quat.y));
    }

    public static float getRotationZFromQuat(Quaternion quat) {
    	return (float)Math.atan2(2 * (quat.w * quat.z + quat.x * quat.y), 1 - 2 * (quat.y * quat.y + quat.z * quat.z));
    }
    
    public static void main(String[] args) {
    	Logger.INSTANCE.open("bla.txt");
    	Quaternion rotation = new Quaternion();
    	Matrix4f mat = MathUtil.rotation(null, PI_DIV2, -PI_DIV2, 0f);
    	rotation.setFromMatrix(mat);
    	rotation.normalise(rotation);
//    	rotation.x = 0.5f;
//    	rotation.y = 0.5f;
//    	rotation.z = 0.5f;
    	Vector3f vecRotation = new Vector3f();
//    	Vector3f vecRotation2 = FromQ(rotation);
    	vecRotation.setX((float) MathUtil.getRotationXFromQuat(rotation));
    	vecRotation.setY((float) MathUtil.getRotationYFromQuat(rotation));
    	vecRotation.setZ((float) MathUtil.getRotationZFromQuat(rotation));
//    	vecRotation.setX((float) vecRotation.x);
//    	vecRotation.setY((float) vecRotation.y);
//    	vecRotation.setZ((float) vecRotation.z);
//    	Logger.INSTANCE.info("Quat ist " + rotation + " Vec ist: " + vecRotation + " oder " + vecRotation2);
	}
    
    //von http://stackoverflow.com/questions/12088610/conversion-between-euler-quaternion-like-in-unity3d-engine
    public static Vector3f rotationVectorFromQuat (Quaternion q1)
    {
        float sqw = q1.w * q1.w;
        float sqx = q1.x * q1.x;
        float sqy = q1.y * q1.y;
        float sqz = q1.z * q1.z;
        float unit = sqx + sqy + sqz + sqw; // if normalised is one, otherwise is correction factor
        float test = q1.x * q1.w - q1.y * q1.z;
        Vector3f v = new Vector3f();

        if (test>0.4995f*unit) { // singularity at north pole
            v.y = (float) (2f * Math.atan2(q1.y, q1.x));
            v.x = MathUtil.PI_DIV2;
            v.z = 0;
            return v;
        }
        if (test<-0.4995f*unit) { // singularity at south pole
            v.y = -2f * (float)Math.atan2 (q1.y, q1.x);
            v.x = -MathUtil.PI_DIV2;
            v.z = 0;
            return v;
        }
        Quaternion q = new Quaternion (q1.w, q1.z, q1.x, q1.y);
        v.y = (float)Math.atan2 (2f * q.x * q.w + 2f * q.y * q.z, 1 - 2f * (q.z * q.z + q.w * q.w));     // Yaw
        v.x = (float)Math.asin (2f * (q.x * q.z - q.w * q.y));                             // Pitch
        v.z = (float)Math.atan2 (2f * q.x * q.y + 2f * q.z * q.w, 1 - 2f * (q.y * q.y + q.z * q.z));      // Roll
        return v;
    }
    
    /**
     * Erzeugt eine Translationsmatrix.
     * @param dest Die Zeilmatrix, kann <tt>null</tt> sein
     * @param translation Der Translationsvektor
     * @return <tt>dest</tt>
     */
    public static Matrix4f translation(Matrix4f dest, Vector3f translation) {
        return translation(dest, translation.x, translation.y, translation.z);
    }
    
    /**
     * Erzeugt eine Translationsmatrix.
     * @param dest Die Zeilmatrix, kann <tt>null</tt> sein
     * @param x Translation entlang der x-Achse
     * @param y Translation entlang der y-Achse
     * @param z Translation entlang der z-Achse
     * @return <tt>dest</tt>
     */
    public static Matrix4f translation(Matrix4f dest, float x, float y, float z) {
        if(dest == null) {
            dest = new Matrix4f();
        }
        dest.setIdentity();
        dest.m30 = x;
        dest.m31 = y;
        dest.m32 = z;
        return dest;
    }
    
    /**
     * Erzeugt eine Skalierungsmatrix.
     * @param dest Die Zeilmatrix, kann <tt>null</tt> sein
     * @param x Skalierung in x-Richtung
     * @param y Skalierung in y-Richtung
     * @param z Skalierung in z-Richtung
     * @return <tt>dest</tt>
     */
    public static Matrix4f scaling(Matrix4f dest, float x, float y, float z) {
        if(dest == null) {
            dest = new Matrix4f();
        }
        dest.setIdentity();
        dest.m00 = x;
        dest.m11 = y;
        dest.m22 = z;
        return dest;
    }
    
    /**
     * Erzeugt eine Skalierungsmatrix.
     * @param dest Die Zeilmatrix, kann <tt>null</tt> sein
     * @param scale Skalierung in alle Richtungen
     * @return <tt>dest</tt>
     */
    public static Matrix4f scaling(Matrix4f dest, float scale) {
        return MathUtil.scaling(dest, scale, scale, scale);
    }
    
    /**
     * Multipliziert beliebig viele Matrizen miteinander.
     * @param dest Zielmatrix, kann <tt>null</tt> sein
     * @param matrices Faktoren der Multiplikation
     * @return Produkt aller Matrizen
     */
    public static Matrix4f mul(Matrix4f dest, Matrix4f... matrices) {
        if(dest == null) {
            dest = new Matrix4f();
        }
        if(matrices.length != 0) {
            dest.load(matrices[0]);
            for(int i=1; i < matrices.length; ++i) {
                mul(dest, dest, matrices[i]);
            }            
        }
        return dest;
    }
    
    /**
     * Multipliziert zwei Matrizen miteinander.
     * @param dest Zielmatrix, kann <tt>null</tt> sein
     * @param left Linker Faktor
     * @param right Rechter Faktor
     * @return <tt>left * right</tt>
     */
    public static Matrix4f mul(Matrix4f dest, Matrix4f left, Matrix4f right) {
        return Matrix4f.mul(left, right, dest);
    }
    
    public static Matrix4f lookTo(Matrix4f dest, Vector3f eyePos, Vector3f viewDir, Vector3f upDir) {
        if(dest == null) {
            dest = new Matrix4f();
        }
        Vector3f sideDir = Vector3f.cross(upDir, viewDir, null);
        Vector3f.cross(viewDir, sideDir, upDir);
        dest.m00 = sideDir.x; dest.m10 = sideDir.y; dest.m20 = sideDir.z; dest.m30 = -Vector3f.dot(eyePos, sideDir);
        dest.m01 = upDir.x;   dest.m11 = upDir.y;   dest.m21 = upDir.z;   dest.m31 = -Vector3f.dot(eyePos, upDir);
        dest.m02 = viewDir.x; dest.m12 = viewDir.y; dest.m22 = viewDir.z; dest.m32 = -Vector3f.dot(eyePos, viewDir);
        dest.m03 = 0.0f;      dest.m13 = 0.0f;      dest.m23 = 0.0f;      dest.m33 = 1.0f;
        return dest;
    }
    
    /**
     * Erstellt eine perspektivische Projektionsmatrix.
     * @param dest Zielmatrix, kann <tt>null</tt> sein
     * @param fov Das Field of view im Bogenmass
     * @param aspectRatio Das Seitenverhaeltnis
     * @param znear Abstand der nahen Clipplane
     * @param zfar Abstand der entfernten Clipplane
     * @return <tt>dest</tt>
     */
    public static Matrix4f perspectiveFov(Matrix4f dest, float fov, float aspectRatio, float znear, float zfar) {
        if(dest == null) {
            dest = new Matrix4f();
        }
        dest.setZero();
        float oneOverTana = 1.0f / tan(0.5f * fov);
        float quot = 1.0f / (zfar - znear);
        dest.m00 = oneOverTana;
        dest.m11 = oneOverTana * aspectRatio;
        //dest.m22 = (zfar + znear) * quot;
        //dest.m32 = - 2.0f * zfar * znear * quot;
        dest.m22 = zfar * quot;
        dest.m32 = -zfar * znear * quot;
        dest.m23 = 1.0f;
        return dest;
    }
    
    /**
     * Wandelt einen Winkel im Winkelmass in einen Winkel im Bogenmass um.
     * @param degree Der Winkel
     * @return <tt>Pi * degree / 180</tt>
     */
    public static float toRadians(float degree) {
        return PI * degree / 180.0f;
    }
    
    public static Matrix4f rotation(Matrix4f dest, Quaternion quat) {
        if(dest == null) {
            dest = new Matrix4f();
        }
        float xx = quat.x * quat.x;
        float xy = quat.x * quat.y;
        float xz = quat.x * quat.z;
        float xw = quat.x * quat.w;
        float yy = quat.y * quat.y;
        float yz = quat.y * quat.z;
        float yw = quat.y * quat.w;
        float zz = quat.z * quat.z;
        float zw = quat.z * quat.w;
        dest.m00 = 1.0f - 2.0f * yy - 2.0f * zz;
        dest.m01 = 2.0f * xy + 2.0f * zw;
        dest.m02 = 2.0f * xz - 2.0f * yw;
        dest.m03 = 0.0f;
        dest.m10 = 2.0f * xy - 2.0f * zw;
        dest.m11 = 1.0f - 2.0f * xx - 2.0f * zz;
        dest.m12 = 2.0f * yz + 2.0f * xw;
        dest.m13 = 0.0f;
        dest.m20 = 2.0f * xz + 2.0f * yw;
        dest.m21 = 2.0f * yz - 2.0f * xw;
        dest.m22 = 1.0f - 2.0f * xx - 2.0f * yy;
        dest.m23 = 0.0f;
        dest.m30 = 0.0f;
        dest.m31 = 0.0f;
        dest.m32 = 0.0f;
        dest.m33 = 1.0f;
        return dest;
    }
    
    /**
     * Erzeugt eine Modelmatrix und optional ihr inverses aus einer Translation
     * und Ausrichtung.
     * @param dest Die Zielmatrix der Modelmatrix, kann <tt>null</tt>sein
     * @param inverseDest Wenn nicht <tt>null</tt> wird die inverse Modelmatrix
     *                    hier eingetragen
     * @param rotation Die Ausrichtung
     * @param translation Die Verschiebung
     * @return <tt>dest</tt>
     */
    public static Matrix4f generateModel(Matrix4f dest, Matrix4f inverseDest, Quaternion rotation, Vector3f translation, float scaling) {
        if(dest == null) {
            dest = new Matrix4f();
        }
        translation(dest, translation);
        rotation(TEMP_MAT, rotation);
        scaling(TEMP_MAT2, scaling);
        Matrix4f.mul(dest, TEMP_MAT2, dest);
        Matrix4f.mul(dest, TEMP_MAT, dest);
        if(inverseDest != null) {
            scaling(TEMP_MAT2, 1.0f / scaling);
            Matrix4f.transpose(TEMP_MAT, TEMP_MAT);
            translation(inverseDest, -translation.x, -translation.y, -translation.z);
            Matrix4f.mul(TEMP_MAT2, inverseDest, inverseDest);
            Matrix4f.mul(TEMP_MAT, inverseDest, inverseDest);
        }
        return dest;
    }
    
    /**
     * Erzeugt eine Rotationsmatrix aus den Roll-, Nick- und Gierwinkeln.
     * @param dest Die Zielmatrix, kann <tt>null</tt> sein
     * @param yaw Gierwinkel im Bogenmass
     * @param pitch Nickwinkel im Bogenmass
     * @param roll Rollwinkel im Bogenmass
     * @return <tt>dest</tt>
     */
    public static Matrix4f rotation(Matrix4f dest, float yaw, float pitch, float roll) {
        dest = rotationZ(dest, roll);
        rotationX(TEMP_MAT2, pitch);
        Matrix4f.mul(TEMP_MAT2, dest, dest);
        rotationY(TEMP_MAT2, yaw);
        Matrix4f.mul(dest, TEMP_MAT2, dest);
        return dest;
    }
    
    /**
     * Erzeugt aus den Roll-, Nick- und Gierwinkeln ein Quaternion.
     * @param dest Die Zielmatrix, kann <tt>null</tt> sein
     * @param yaw Gierwinkel im Bogenmass
     * @param pitch Nickwinkel im Bogenmass
     * @param roll Rollwinkel im Bogenmass
     * @return <tt>dest</tt>
     */
    public static Quaternion quaternionFromYPR(Quaternion dest, float yaw, float pitch, float roll) {
        if(dest == null) {
            dest = new Quaternion();
        }
        dest.setFromMatrix(rotation(TEMP_MAT, yaw, pitch, roll));
        return dest;
    }
    
//    /**
//     * Ueberfuehrt ein Ode Quaternion in eine LWJGL Version.
//     * @param dest Das Zielquaternion, kann <tt>null</tt> sein
//     * @param quat Das Ursprungsquaternion
//     * @return <tt>dest</tt>
//     */
//    public static Quaternion ode2lwjgl(Quaternion dest, DQuaternionC quat) {
//        if(dest == null) {
//            dest = new Quaternion();
//        }
//        dest.set((float)quat.get0(), (float)quat.get1(), (float)quat.get2(), (float)quat.get3());
//        return dest;
//    }
//    
//    /**
//     * Ueberfuehrt ein LWJGL Quaternion in eine Ode Version.
//     * @param dest Das Zielquaternion, kann <tt>null</tt> sein
//     * @param quat Das Ursprungsquaternion
//     * @return <tt>dest</tt>
//     */
//    public static DQuaternion lwjgl2ode(DQuaternion dest, Quaternion quat) {
//        if(dest == null) {
//            dest = new DQuaternion();
//        }
//        dest.set(quat.x, quat.y, quat.z, quat.w);
//        return dest;
//    }
    
    /**
     * Ueberfuehrt einen vecmath Vektor in eine LWJGL Version.
     * @param dest Der Zielvektor, kann <tt>null</tt> sein
     * @param vec Der Ursprungsvektor
     * @return <tt>dest</tt>
     */
    public static Vector3f vecmath2lwjgl(Vector3f dest, javax.vecmath.Vector3f vec) {
        if(dest == null) {
            dest = new Vector3f();
        }
        dest.set(vec.x, vec.y, vec.z);
        return dest;
    }
    
    /**
     * Ueberfuehrt einen LWJGL Vektor in eine vecmath Version.
     * @param dest Der Zielvektor, kann <tt>null</tt> sein
     * @param vec Der Ursprungsvektor
     * @return <tt>dest</tt>
     */
    public static javax.vecmath.Vector3f lwjgl2vecmath(javax.vecmath.Vector3f dest, Vector3f vec) {
        if(dest == null) {
            dest = new javax.vecmath.Vector3f();
        }
        dest.set(vec.x, vec.y, vec.z);
        return dest;
    }
    
    public static javax.vecmath.Matrix4f lwjgl2vecmath(javax.vecmath.Matrix4f dest, Matrix4f mat) {
        if(dest == null) {
            dest = new javax.vecmath.Matrix4f();
        }
        dest.m00 = mat.m00; dest.m01 = mat.m10; dest.m02 = mat.m20; dest.m03 = mat.m30;
        dest.m10 = mat.m01; dest.m11 = mat.m11; dest.m12 = mat.m21; dest.m13 = mat.m31;
        dest.m20 = mat.m02; dest.m21 = mat.m12; dest.m22 = mat.m22; dest.m23 = mat.m32;
        dest.m30 = mat.m03; dest.m31 = mat.m13; dest.m32 = mat.m23; dest.m33 = mat.m33;
        return dest;
    }
    
    public static Matrix4f vecmath2lwjgl(Matrix4f dest, javax.vecmath.Matrix4f mat) {
        if(dest == null) {
            dest = new Matrix4f();
        }
        dest.m00 = mat.m00; dest.m01 = mat.m10; dest.m02 = mat.m20; dest.m03 = mat.m30;
        dest.m10 = mat.m01; dest.m11 = mat.m11; dest.m12 = mat.m21; dest.m13 = mat.m31;
        dest.m20 = mat.m02; dest.m21 = mat.m12; dest.m22 = mat.m22; dest.m23 = mat.m32;
        dest.m30 = mat.m03; dest.m31 = mat.m13; dest.m32 = mat.m23; dest.m33 = mat.m33;
        return dest;
    }
    
    /**
     * Ruft die Methode <tt>toRadians()</tt> fuer jeden Eintrag eines Vektors
     * auf.
     * @param srcAndDest Der Vektor
     * @return <tt>srcAndDest</tt>
     */
    public static Vector3f toRadians(Vector3f srcAndDest) {
        srcAndDest.x = toRadians(srcAndDest.x);
        srcAndDest.y = toRadians(srcAndDest.y);
        srcAndDest.z = toRadians(srcAndDest.z);
        return srcAndDest;
    }
    
    public static Quaternion vecmath2lwjgl(Quaternion dest, Quat4f quat) {
        if(dest == null) {
            dest = new Quaternion();
        }
        dest.set(quat.x, quat.y, quat.z, quat.w);
        return dest;
    }
    
    public static Quat4f lwjgl2vecmath(Quat4f dest,  Quaternion quat) {
        if(dest == null) {
            dest = new Quat4f();
        }
        dest.set(quat.x, quat.y, quat.z, quat.w);
        return dest;
    }
    
    public static void split(Matrix4f src, Vector3f position, Quaternion rotation) {
        position.x = src.m30;
        position.y = src.m31;
        position.z = src.m32;
        rotation.setFromMatrix(src);
    }
    
    public static float lerp(float val1, float val2, float t) {
        return val1 + t * (val2 - val1);
    }
    
    public static Vector4f lerp(Vector4f dest, Vector4f val1, Vector4f val2, float t) {
        if(dest == null) {
            dest = new Vector4f();
        }
        dest.x = lerp(val1.x, val2.x, t);
        dest.y = lerp(val1.y, val2.y, t);
        dest.z = lerp(val1.z, val2.z, t);
        dest.w = lerp(val1.w, val2.w, t);
        return dest;
    }
    
    public static Vector3f lerp(Vector3f dest, Vector3f val1, Vector3f val2, float t) {
        if(dest == null) {
            dest = new Vector3f();
        }
        dest.x = lerp(val1.x, val2.x, t);
        dest.y = lerp(val1.y, val2.y, t);
        dest.z = lerp(val1.z, val2.z, t);
        return dest;
    }
    
    public static Transform lwjgl2jbullet(Transform dest, Vector3f position, Quaternion rotation) {
        if(dest == null) {
            dest = new Transform();
        }
        lwjgl2vecmath(dest.origin, position);
        dest.setRotation(lwjgl2vecmath(null, rotation));
        return dest;
    }
    
    /**
     * Wandelt eine Anzahl von Bytes in einen besser lesbaren String um.
     * @param bytes Die Bytes
     * @return Der String
     */
    public static String byteToString(long bytes) {
        String unit = "B";
        float lSize = bytes;
        if(lSize > 2048.0f) {
            lSize /= 1024.0f;
            unit = "KB";
        }
        if(lSize > 2048.0f) {
            lSize /= 1024.0f;
            unit = "MB";
        }
        if(lSize > 2048.0f) {
            lSize /= 1024.0f;
            unit = "GB";
        }
        return String.format("%.1f %s", lSize, unit);
    }

    public static float sqrt(float a) {
        return (float)Math.sqrt(a);
    }
    
    public static Vector3f transformCoord(Vector3f dest, Matrix4f left, Vector3f right) {
        if(dest == null) {
            dest = new Vector3f();
        }
        TEMP_VEC4_A.set(right.x, right.y, right.z, 1.0f);
        transformCoord(TEMP_VEC4_A, left, TEMP_VEC4_A);
        dest.set(TEMP_VEC4_A);
        return dest;
    }
    
    public static Vector4f transformCoord(Vector4f dest, Matrix4f left, Vector4f right) {
        dest = Matrix4f.transform(left, right, dest);
        float oneOverW = 1.0f / dest.w;
        dest.x *= oneOverW;
        dest.y *= oneOverW;
        dest.z *= oneOverW;
        dest.w = 1.0f;
        return dest;
    }
    
    public static Vector4f createPlane(Vector4f dest, Vector3f a, Vector3f b, Vector3f c) {
        if(dest == null) {
            dest = new Vector4f();
        }
        Vector3f ba = Vector3f.sub(a, b, TEMP_VEC3_A);
        Vector3f bc = Vector3f.sub(c, b, TEMP_VEC3_B);
        Vector3f normal = Vector3f.cross(ba, bc, TEMP_VEC3_C);
        normal.normalise();
        dest.set(normal.x, normal.y, normal.z, -Vector3f.dot(b, normal));
        return dest;
    }
}
