/*
 * Copyright (C) 2013 nico3000 - All Rights Reserved
 */
package util;

import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Quaternion;
import org.lwjgl.util.vector.Vector3f;
import org.w3c.dom.Node;


/**
 * Eine Pose repraesentiert die Ausrichtung und Position eines Objekts.
 * @author nico3000
 */
public class Pose {
    /**
     * Die Position des Objekts.
     */
    private final Vector3f position = new Vector3f();
    /**
     * Die Ausrichtung des Objekts.
     */
    private final Quaternion rotation = new Quaternion();
    /**
     * Die SKalierung des Actors.
     */
    private float scale = 1.0f;
    /**
     * Die Modelmatrix.
     */
    private final Matrix4f model = Matrix4f.setIdentity(new Matrix4f());
    /**
     * Die inverse Modelmatrix.
     */
    private final Matrix4f modelInv = Matrix4f.setIdentity(new Matrix4f());
    /**
     * Wird <tt>true</tt> wenn sich die Pose geaendert hat.
     */
    private boolean changed = true;
    /**
     * Actor, der mit dieser Pose verknuepft ist, sofern vorhanden.
     */
    private final int actorId;
    
    /**
     * Konstruktor.
     * @param actorId Der Actor, der mit dieser Pose verknuepft ist
     */
    public Pose(int actorId) {
        this.actorId = actorId;
    }
    
    /**
     * Setzt Position und Rotation.
     * @param position Die neue Position
     * @param rotation Die neue Rotation
     */
    public void set(Vector3f position, Quaternion rotation) {
        this.position.set(position);
        this.rotation.set(rotation);
        this.setChanged();
    }
    
    /**
     * Setzt die Ausrichtung und Position anhand eines XML Knoten.
     * @param data Der erste Kindnoten des Knotens, der die Daten enthaelt
     */
    public void setFromXml(Node data) {
        this.position.set(0.0f, 0.0f, 0.0f);
        Vector3f ypr = new Vector3f();
        while(data != null) {
            switch(data.getNodeName()) {
                case "Position":
                    XMLUtil.getAttributesAsVector(data, "x", "y", "z", this.position);
                    break;
                case "Rotation":
                    XMLUtil.getAttributesAsVector(data, "yaw", "pitch", "roll", ypr);
                    MathUtil.toRadians(ypr);                    
                    break;
                case "Scale":
                    this.scale = XMLUtil.getAttributeAsFloat(data, "value", 1.0f);
                    break;
            }
            data = data.getNextSibling();
        }
        this.setRotation(ypr.x, ypr.y, ypr.z);
        this.setChanged();
    }
    
    /**
     * Setzt die Position des zugehoerigen Actors.
     * @param x Koordinate
     * @param y Koordinate
     * @param z Koordinate 
     */
    public void setPosition(float x, float y, float z) {
        this.position.set(x, y, z);
        this.setChanged();
    }
    
    /**
     * Setzt die Position des zugehoerigen Actors.
     * @param position Die Position
     */
    public void setPosition(Vector3f position) {
        this.setPosition(position.x, position.y, position.z);
    }  

    /**
     * Bewegt die Kamera relativ zu ihrer aktuellen Ausrichtung.
     * @param forwardBackward Bewegung nach vorn (positiv) oder hinten (negativ)
     * @param rightLeft Bewegung nach rechts (positiv) oder links (negativ)
     * @param upDown Bewegung nach oben (positiv) oder unten (negativ)
     * @param globalUpDown Wenn <tt>true</tt>, bewegt sich der Actor entlang der
     *                     globalen y-Achse nach oben bzw. unten
     */
    public void moveRelative(float forwardBackward, float rightLeft, float upDown, boolean globalUpDown) {
        this.updateModel();
        this.position.x += forwardBackward * this.model.m20 + rightLeft * this.model.m00;
        this.position.y += forwardBackward * this.model.m21 + rightLeft * this.model.m01;
        this.position.z += forwardBackward * this.model.m22 + rightLeft * this.model.m02;
        if(globalUpDown) {
            this.position.y += upDown;
        } else {
            this.position.x += upDown * this.model.m10;
            this.position.y += upDown * this.model.m11;
            this.position.z += upDown * this.model.m12;
        }
        this.setChanged();
    }
    
    /**
     * Setzt die Rotation des zugehoerigen Actors.
     * @param rotation Die Rotation
     */
    public void setRotation(Quaternion rotation) {
        this.setRotation(rotation.x, rotation.y, rotation.z, rotation.w);
    }

    /**
     * Setzt die Rotation des zugehoerigen Actors.
     * @param x Erster Eintrag des Quaternions
     * @param y Zweiter Eintrag des Quaternions
     * @param z Dritter Eintrag des Quaternions
     * @param w Vierter Eintrag des Quaternions
     */
    public void setRotation(float x, float y, float z, float w) {
        this.rotation.set(x, y, z, w);
        this.setChanged();
    }
    
    /**
     * Setzt die Rotation anhand von Gier-, Nick- und Rollwinkel.
     * @param yaw Der Gierwinkel
     * @param pitch Der Nickwinkel
     * @param roll Der Rollwinkel
     */
    public void setRotation(float yaw, float pitch, float roll) {
        this.rotation.setFromMatrix(MathUtil.rotation(null, yaw, pitch, roll));
        this.setChanged();
    }
    
    /**
     * Gibt die Rotation als Vector3 zurueck
     */
    public Vector3f getRotationVector() {
    	return MathUtil.rotationVectorFromQuat(rotation);
    }

    /**
     * Setzt eine neue Skalierung des zugehoerigen Actors.
     * @param scale Die neue Skalierung
     */
    public void setScale(float scale) {
        this.scale = scale;
        this.setChanged();
    }

    /**
     * Position des Objekts.
     * @return Die Position
     */
    public Vector3f getPosition() {
        return this.position;
    }

    /**
     * Rotation des Objekts.
     * @return Die Rotation als Quaternion
     */
    public Quaternion getRotation() {
        return this.rotation;
    }

    /**
     * Die Skalierung des zugehoerigen Actors.
     * @return Die Skalierung
     */
    public float getScale() {
        return this.scale;
    }
    
    /**
     * Setzt den Zustand der Pose auf geaendert. Der naechtse Aufruf von
     * <tt>updateModel()</tt> generiert dann die Modelmatrix neu und gibt
     * <tt>true</tt> zurueck.
     */
    public void setChanged() {      
        this.changed = true;
    }
    
    /**
     * Generiert die Rueckgabewerte von <tt>getModel()</tt> und
     * <tt>getModelInv()</tt> neu. Sollte aufgerufen, nachdem man alle
     * Aenderungen mittels <tt>setPosition()</tt> und <tt>setRotation()</tt>
     * vorgenommen hat.
     * @return <tt>true</tt>, wenn sich etwas geaendert hat
     */    
    public boolean updateModel() {
        if(this.changed) {
            MathUtil.generateModel(this.model, this.modelInv, this.rotation, this.position, this.scale);
            this.changed = false;
            return true;
        } else {
            return false;
        }
    }
    
    /**
     * Liefert die Modelmatrix.
     * @return Die Modelmatrix
     */
    public Matrix4f getModel() {
        return this.model;
    }
    
    /**
     * Liefert die inverse Modelmatrix.
     * @return Die inverse Modelmatrix
     */
    public Matrix4f getModelInv() {
        return this.modelInv;
    }
    
    /**
     * Liefert den Vektor der Laenge 1, der in Richtung der lokalen x-Achse
     * zeigt.
     * @param dest Der Zielvektor, darf <tt>null</tt> sein
     * @return <tt>dest</tt>
     */
    public Vector3f getLocalXAxis(Vector3f dest) {
        if(dest == null) {
            dest = new Vector3f();
        }
        dest.x = this.model.m00;
        dest.y = this.model.m01;
        dest.z = this.model.m02;
        return dest;
    }
    
    /**
     * Liefert den Vektor der Laenge 1, der in Richtung der lokalen y-Achse
     * zeigt.
     * @param dest Der Zielvektor, darf <tt>null</tt> sein
     * @return <tt>dest</tt>
     */
    public Vector3f getLocalYAxis(Vector3f dest) {
        if(dest == null) {
            dest = new Vector3f();
        }
        dest.x = this.model.m10;
        dest.y = this.model.m11;
        dest.z = this.model.m12;
        return dest;
    }
    
    /**
     * Liefert den Vektor der Laenge 1, der in Richtung der lokalen z-Achse
     * zeigt.
     * @param dest Der Zielvektor, darf <tt>null</tt> sein
     * @return <tt>dest</tt>
     */
    public Vector3f getLocalZAxis(Vector3f dest) {
        if(dest == null) {
            dest = new Vector3f();
        }
        dest.x = this.model.m20;
        dest.y = this.model.m21;
        dest.z = this.model.m22;
        return dest;
    }

    /**
     * Liefert die ID des Actors.
     * @return Die ID des Actors
     */
    public int getActorId() {
        return actorId;
    }
}
